"""
Test tee module
===============

"""

import unittest
import os
from shutil import rmtree
import sys

from ..tee import MultiFile

from ...io.redirect_stdout import stdout_redirected


class TestUtil(unittest.TestCase):
    """Test fluiddyn.util.util module."""

    @classmethod
    def setUpClass(cls):
        cls._work_dir = "test_fluiddyn_io_tee"
        if not os.path.exists(cls._work_dir):
            os.mkdir(cls._work_dir)

        os.chdir(cls._work_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test_tee(self):

        mf = MultiFile([sys.stdout])
        with stdout_redirected():
            mf.write("")


if __name__ == "__main__":
    unittest.main()



"""
Test hdf5 module
================

"""

import unittest
import os
from shutil import rmtree

import numpy as np
import h5py

from .. import util
from ..terminal_colors import print_fail, print_warning
from ..userconfig import load_user_conf_files

from ...io.redirect_stdout import stdout_redirected


def expectedFailureIf(condition):
    if condition:
        return unittest.expectedFailure
    else:
        return lambda func: func


class MyObject:
    def __init__(self, str_path=None, *args, **kwargs):
        pass

    def save(self, path):
        with h5py.File(path, "w") as f:
            f.attrs["class_name"] = "MyObject"
            f.attrs["module_name"] = "fluiddyn.util.test.test_util"


class TestUtil(unittest.TestCase):
    """Test fluiddyn.util.util module."""

    @classmethod
    def setUpClass(cls):
        cls._work_dir = "test_fluiddyn_util_util"
        if not os.path.exists(cls._work_dir):
            os.mkdir(cls._work_dir)

        os.chdir(cls._work_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test_util(self):

        util.import_class("fluiddyn.output.figs", "Figures")

        util.time_as_str(decimal=1)

        util.is_run_from_ipython()
        util.is_run_from_jupyter()
        with stdout_redirected():
            util.print_memory_usage("test")
            util.print_size_in_Mo(np.arange(4))
            util.print_size_in_Mo(np.arange(4), string="test")
            print_fail("")
            print_warning("")

        with util.print_options():
            pass

        util.config_logging()
        load_user_conf_files()

    def test_copy_me(self):
        # Relative path before chdir are supplied with Python 2
        util.copy_me_in(os.curdir)

    def test_mod_date(self):
        # Relative path before chdir are supplied with Python 2
        util.modification_date(os.path.dirname(__file__))

    def test_create_object(self):
        o = MyObject()
        o.save("myobject.h5")
        util.create_object_from_file("object")


if __name__ == "__main__":
    unittest.main()



"""Container for parameters (:mod:`fluiddyn.util.paramcontainer`)
=================================================================

.. currentmodule:: fluiddyn.util.paramcontainer

Provides:

.. autoclass:: ParamContainer
   :members:
   :private-members:

"""

from typing import Dict, Any
import os
from copy import deepcopy

from io import open

try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree

from ast import literal_eval
import six
import re

import numpy as np

try:
    import h5py
    from fluiddyn.io.hdf5 import H5File
except ImportError:
    from warnings import warn

    warn(
        "Cannot import h5py. Loading from and writing to HDF5 files will not"
        "work",
        ImportWarning,
    )

from fluiddyn.io import Path
from fluiddyn.util.xmltotext import produce_text_element


def _as_str(value):
    if isinstance(value, Path):
        return str(value)

    elif not isinstance(value, str):
        return repr(value)

    else:
        return value


def _as_value(value):
    if value.startswith("array("):
        return eval(value)

    if "\t" in value:
        return value

    try:
        return literal_eval(value)

    except (SyntaxError, ValueError):
        return value


def sanitize_for_json(d: Dict[str, Any]) -> Dict[str, Any]:
    """JSON rendering requires values in the dictionary to be a basic type.

    That is the allowed types are ``{str, int, float, bool, None}``. If not,
    sanitize beforehand.

    """

    def is_not_allowed(value):
        allowed_types = (str, int, float, bool)
        return not (value is None or isinstance(value, allowed_types))

    for key, value in d.items():
        if isinstance(value, dict):
            d[key] = sanitize_for_json(value)  # recurse

        elif isinstance(value, (list, tuple)):
            d[key] = [_as_str(v) if is_not_allowed(v) else v for v in value]

        elif is_not_allowed(value):
            d[key] = _as_str(value)

    return d


class ParamContainer:
    """Structured container of values.

    The objects ParamContainer can be used as containers of
    parameters. They can be printed as xml text.

    They are used to contain parameters that can be modified by the
    user, for example in this way::

      >>> params = ParamContainer(tag='params')

      >>> params._set_attribs({'a0': 1, 'a1': 1})

      >>> params._set_attrib('a2', 1)

      >>> params._print_as_xml()
      <params a1="1" a0="1" a2="1"/>

      >>> params._set_child('child0', {'a0': 2, 'a1': 2})

      >>> params.child0.a0 = 3

      >>> params._print_as_xml()
      <params a1="1" a0="1" a2="1">
      <child0 a1="2" a0="3"/>
      </params>

    Here, ``params.child0`` is another ParamContainer.

    An interesting feature of the ParamContainer objects is that if one
    uses a non-existing parameter, an AttributeError is raised::

      >>> params.a3 = 3

      -------------------------------------------------------------------------
      AttributeError                          Traceback (most recent call last)
      <ipython-input-9-a91118d8b23e> in <module>()
      ----> 1 params.child1.a0 = 3

      AttributeError: a3 is not already set in params.
      The attributes are: set(['a1', 'a0', 'a2'])
      To set a new attribute, use _set_attrib or _set_attribs.

    Note that for a parameter container, it is much better to raise an
    error rather than just add an unused parameter!

    A ParamContainer object can be saved in xml and in hdf5 and then
    reloaded from the file::

      >>> params._save_as_xml()

      >>> params_loaded = ParamContainer(path_file=params._tag + '.xml')

      >>> assert params_loaded == params

    Parameters
    ----------
    (for the __init__ method)

    tag : (None) str
        A tag for the root container.

    attribs : (None) dict
        Some attributes.

    path_file : (None) str
        Path of a file (xml or hdf5).

    elemxml : (None) xml.etree.ElementTree.Element
        An xml element.

    hdf5_object : (None) file
        An open hdf5 file.

    """

    def __init__(
        self,
        tag=None,
        attribs=None,
        path_file=None,
        elemxml=None,
        hdf5_object=None,
        doc="",
        parent=None,
    ):

        self._set_internal_attr("_key_attribs", list())
        self._set_internal_attr("_tag_children", list())
        self._set_internal_attr("_parent", parent)
        self._set_doc(doc)

        if path_file is not None:
            path_file = str(path_file)
            self._set_internal_attr("_path_file", path_file)
            if path_file.endswith(".xml"):
                self._load_from_xml_file(path_file)
            elif path_file.endswith(".h5"):
                self._load_from_hdf5_file(path_file)
        elif elemxml is not None:
            self._load_from_elemxml(elemxml)
        elif hdf5_object is not None:
            self._load_from_hdf5_object(hdf5_object)
        elif tag is not None:
            self._set_internal_attr("_tag", tag)
        else:
            raise ValueError(
                "To create an empty ParamContainer, " "a tag has to be provided."
            )

        if attribs is not None:
            self._set_attribs(attribs)

    def __setattr__(self, key, value):
        if key in self._key_attribs:
            self._set_internal_attr(key, value)
        elif key in self._tag_children:
            if not isinstance(value, ParamContainer):
                raise AttributeError(key + " is a tag of a child.")

        else:
            raise AttributeError(
                key
                + " is not already set in "
                + self._tag
                + ".\nThe attributes are: "
                + str(self._key_attribs)
                + "\nTo set a new attribute, use _set_attrib or _set_attribs."
            )

    def __getattr__(self, attr):
        if attr.startswith("_"):
            raise AttributeError(
                f"{self.__class__} object has no attribute {attr}"
            )

        else:
            raise AttributeError(
                attr
                + " is not an attribute of "
                + self._tag
                + ".\nThe attributes are: "
                + str(self._key_attribs)
                + "\nThe children are: "
                + str(self._tag_children)
            )

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, key):
        return self.__getattribute__(key)

    def _set_internal_attr(self, key, value):
        self.__dict__[key] = value

    def _set_doc(self, doc):
        self._set_internal_attr("_doc", doc)

    def _contains_doc(self):
        if len(self._doc) > 0:
            return True

        for tag in self._tag_children:
            if self[tag]._contains_doc():
                return True

        return False

    def _get_formatted_doc(self):
        full_tag = self._make_full_tag()
        txt = "Documentation for " + full_tag

        nb_points = full_tag.count(".")
        if nb_points == 0:
            char = "="
        elif nb_points == 1:
            char = "-"
        elif nb_points == 2:
            char = "~"
        else:
            char = "^"

        txt += "\n" + char * len(txt) + "\n"

        doc = self._doc.strip()

        if len(doc) == 0:
            return txt + "\n"

        if len(doc) > 0:
            txt += "\n"

        txt += doc

        if len(doc) > 0:
            txt += "\n"

        return txt

    def _print_doc(self):
        print(self._get_formatted_doc())

    def _get_formatted_docs(self):
        txt = self._get_formatted_doc()
        for tag in self._tag_children:
            if not txt.endswith("\n\n"):
                txt += "\n"
            txt += self[tag]._get_formatted_docs()

        return txt

    def _print_docs(self):
        print(self._get_formatted_docs())

    def _make_full_tag(self):
        if self._parent is None:
            return self._tag

        else:
            return self._parent._make_full_tag() + "." + self._tag

    def _set_attrib(self, key, value):
        """Add an attribute to the container."""
        self.__dict__[key] = value
        self._key_attribs.append(key)

    def _set_attribs(self, d):
        """Add the attributes to the container."""
        for k, v in list(d.items()):
            self._set_attrib(k, v)

    def _get_key_attribs(self):
        self._key_attribs.sort()
        return self._key_attribs

    def _set_child(self, tag, attribs=None):
        """Add a child (of the same class) to the container."""
        if tag in self.__dict__:
            raise ValueError(f'The tag "{tag}" is already used.')

        self.__dict__[tag] = self.__class__(tag=tag, attribs=attribs, parent=self)
        self._tag_children.append(tag)

    def _set_as_child(self, child, change_parent=True):
        """Associate a ParamContainer as a child."""
        if not isinstance(child, ParamContainer):
            raise ValueError("child should be a ParamContainer instance.")

        if child._tag in self.__dict__:
            raise ValueError(f'The tag "{child._tag}" is already used.')

        self.__dict__[child._tag] = child
        if change_parent:
            child._set_internal_attr("_parent", self)
        self._tag_children.append(child._tag)

    def _make_dict_attribs(self):
        d = {}
        for k in self._key_attribs:
            d[k] = self.__dict__[k]
        return d

    def _make_dict(self):
        self._key_attribs.sort()
        d = {
            "tag": self._tag,
            "key_attribs": self._key_attribs,
            "tag_children": self._tag_children,
        }

        attribs = d["attribs"] = []
        for k in self._key_attribs:
            attribs.append(self.__dict__[k])

        children = d["children"] = []
        for k in self._tag_children:
            children.append(self.__dict__[k]._make_dict())

        return d

    def _make_dict_tree(self):
        """A tree-like nested dictionary, including attributes and children."""
        d = self._make_dict_attribs()
        for k in self._tag_children:
            d[k] = self.__dict__[k]._make_dict_tree()
        return d

    def __eq__(self, other):
        return self._make_dict() == other._make_dict()

    def __sub__(self, other):
        """Subtract and return an ParamContainer to understand differences."""
        if self == other:
            return None

        this = deepcopy(self)

        for k in self._key_attribs:
            if self.__dict__[k] == other.__dict__[k]:
                this.__dict__.pop(k)
                this._key_attribs.remove(k)

        for k in self._tag_children:
            this.__dict__[k] = self.__dict__[k] - other.__dict__[k]
            if this.__dict__[k] is None:
                this.__dict__.pop(k)
                this._tag_children.remove(k)

        return this

    def _make_element_xml(self, parentxml=None):
        if parentxml is None:
            elemxml = etree.Element(self._tag)
        else:
            elemxml = etree.SubElement(parentxml, self._tag)

        self._key_attribs.sort()
        for key in self._key_attribs:
            elemxml.attrib[key] = _as_str(self.__dict__[key])

        if hasattr(self, "_value_text"):
            elemxml.text = str(self._value_text)

        for key in self._tag_children:
            child = self.__dict__[key]
            child._make_element_xml(elemxml)

        return elemxml

    def _make_xml_text(self):
        """Produce and return a xml text representing the container."""
        elemxml = self._make_element_xml()
        return produce_text_element(elemxml)

    def _print_as_xml(self):
        """Print the xml text representing the container."""
        print(self._make_xml_text())

    def __repr__(self):
        return super().__repr__() + "\n\n" + self._make_xml_text()

    def _repr_json_(self):
        data = self._make_dict_tree()
        data = sanitize_for_json(data)

        this_id = hex(id(self))
        metadata = {
            "root": f"{self._tag} of {self.__class__} at {this_id}",
            "expanded": True,
        }
        return data, metadata

    def _load_from_xml_file(self, path_file):
        tree = etree.parse(path_file)
        elemxml = tree.getroot()
        self._load_from_elemxml(elemxml)

    def _load_from_elemxml(self, elemxml):
        self._set_internal_attr("_tag", elemxml.tag)

        text = elemxml.text
        if text is not None:
            v = text.strip()
            if len(v) > 0:
                self._set_internal_attr("_value_text", _as_value(v))

        for k, v in list(elemxml.attrib.items()):
            self._set_attrib(k, _as_value(v))

        self._key_attribs.sort()

        tags_multiple = []

        for childxml in elemxml:
            tag = childxml.tag

            children = [c for c in elemxml if c.tag == tag]

            if len(children) > 1 or tag in tags_multiple:
                if tag not in tags_multiple:
                    tags_multiple.append(tag)

                if len(childxml.attrib) == 1:
                    k = list(childxml.attrib.keys())[0]
                    tag += "_" + str(childxml.attrib.pop(k))
                    childxml.tag = tag

            self._set_internal_attr(tag, self.__class__(elemxml=childxml))
            self._tag_children.append(tag)

    def _save_as_xml(self, path_file=None, comment=None, find_new_name=False):
        """Save the xml text in a file."""
        if path_file is None:
            path_file = self._tag + ".xml"

        if os.path.exists(path_file):
            if not find_new_name:
                raise ValueError(f"The file {path_file} already exists.")

            else:
                base = path_file.split(".xml")[0]
                i = 1
                while os.path.exists(base + f"_{i}.xml"):
                    i += 1
                path_file = base + f"_{i}.xml"

        with open(path_file, "w", encoding="utf-8") as f:
            if comment is not None:
                try:
                    comment = comment.decode("utf-8")
                except (AttributeError, UnicodeEncodeError):
                    pass
                f.write("<!--\n" + comment + "\n-->\n")
            f.write(self._make_xml_text())
        return path_file

    def _save_as_hdf5(
        self,
        path_file=None,
        hdf5_object=None,
        hdf5_parent=None,
        invalid_netcdf=False,
    ):
        """Save in a hdf5 file."""

        if hdf5_parent is not None:
            hdf5_object = hdf5_parent.create_group(self._tag)

        if hdf5_object is None:
            if path_file is None:
                path_file = ""
            if not path_file.endswith(".h5"):
                path_file = os.path.join(path_file, self._tag + ".h5")
            with H5File(path_file, "w") as f:
                try:
                    tag = self._tag.encode("utf8")
                except AttributeError:
                    tag = self._tag
                f.attrs["_tag"] = tag
                self._save_as_hdf5(hdf5_object=f)
        elif path_file is None:
            for key in self._key_attribs:
                value = self.__dict__[key]
                if value is None:
                    value = "None"
                try:
                    value = value.encode("utf8")
                except (AttributeError, UnicodeDecodeError):
                    pass

                if isinstance(value, (list, tuple)):
                    value_tmp = []
                    for v in value:
                        try:
                            v_tmp = v.encode("utf8")
                        except AttributeError:
                            v_tmp = v
                        value_tmp.append(v_tmp)
                    if isinstance(value, tuple):
                        value = tuple(value_tmp)
                    else:
                        value = value_tmp

                if not invalid_netcdf and isinstance(value, bool):
                    value = int(value)

                if isinstance(value, Path):
                    value = str(value)

                try:
                    hdf5_object.attrs[key] = value
                except TypeError:
                    print(key, value, type(value))
                    raise

            for key in self._tag_children:
                group = hdf5_object.create_group(key)
                self.__dict__[key]._save_as_hdf5(hdf5_object=group)
        else:
            raise ValueError(
                "If hdf5_object is not None," "path_file should be None."
            )

    def _load_from_hdf5_file(self, path_file):
        with H5File(path_file, "r") as f:
            self._load_from_hdf5_object(f)

    def _load_from_hdf5_object(self, hdf5_object):

        attrs = dict(hdf5_object.attrs)

        for k, v in list(attrs.items()):
            if isinstance(v, np.integer):
                attrs[k] = int(v)
                continue

            try:
                attrs[k] = v.decode("utf8")
            except AttributeError:
                pass

            if isinstance(v, np.ndarray) and v.dtype.kind in ("S", "U"):
                attrs[k] = list(v.astype(np.unicode))

        tag = hdf5_object.name.split("/")[-1]

        if tag == "":
            try:
                tag = hdf5_object.attrs["_tag"]
                attrs.pop("_tag")
            except KeyError:
                tag = "root_file"

        try:
            tag = tag.decode("utf8")
        except AttributeError:
            pass

        self._set_internal_attr("_tag", tag)

        # detect None attributes
        for k, v in list(attrs.items()):
            if isinstance(v, str) and v == "None":
                attrs[k] = None

        for key in list(attrs.keys()):
            if " " in key:
                attrs[key.replace(" ", "_")] = attrs.pop(key)

        self._set_attribs(attrs)
        for tag in list(hdf5_object.keys()):
            if isinstance(hdf5_object[tag], h5py.Dataset):
                value = hdf5_object[tag][...]
                self._set_attrib(tag, value)

            elif isinstance(hdf5_object[tag], h5py.Group):
                self.__dict__[tag] = self.__class__(hdf5_object=hdf5_object[tag])
                self._tag_children.append(tag)

    def _modif_from_other_params(self, params):
        """Modify the object with another similar container."""
        for key in params._key_attribs:
            try:
                self[key] = params[key]
            except AttributeError:
                pass
        for tag_child in params._tag_children:
            try:
                self[tag_child]._modif_from_other_params(params[tag_child])
            except AttributeError:
                pass


def tidy_container(cont):
    """Modify the names in a ParamContainer and its organization.

    """
    newtag = convert_capword_to_lowercaseunderscore(cont._tag)
    cont._set_internal_attr("_tag", newtag)

    for i, oldtag in enumerate(cont._tag_children):
        newtag = convert_capword_to_lowercaseunderscore(oldtag)
        cont._tag_children[i] = newtag
        cont.__dict__[newtag] = cont.__dict__.pop(oldtag)

        if isinstance(cont.__dict__[newtag], str):
            print(cont.__dict__[newtag])
        tidy_container(cont.__dict__[newtag])

    tag_child_to_remove = []
    for tag in cont._tag_children:
        child = cont.__dict__[tag]
        if len(child._tag_children) == 0 and len(child._key_attribs) == 0:
            try:
                value_text = child._value_text
            except AttributeError:
                value_text = ""
            cont.__dict__.pop(tag)
            tag_child_to_remove.append(tag)
            cont._set_attrib(tag, value_text)

    for tag in tag_child_to_remove:
        cont._tag_children.remove(tag)


def convert_capword_to_lowercaseunderscore(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


if __name__ == "__main__":
    # params = ParamContainer(tag='params')

    # params._set_attrib('a0', 1)
    # params._set_attribs({'a1': 1, 'a2': 1})

    # params._set_child('child0', {'a0': 2, 'a1': 2})

    # params2 = ParamContainer(tag='params')

    # params2._set_attribs({'a1': 1, 'a2': 1})
    # params2._set_attrib('a0', 1)

    # params2._set_child('child0', {'a0': 2, 'a1': 2})

    # assert params == params2

    # c1 = ParamContainer(tag='child1')
    # c1._set_attrib('a0', 10)

    # params._set_as_child(c1)

    # print(params)

    # assert params != params2

    params = ParamContainer(tag="params")
    params._set_attribs({"a0": 1, "a1": 1})
    params._set_attrib("a2", 1)

    params._print_as_xml()

    params._set_child("child0", {"a0": 2, "a1": 2})
    params.child0.a0 = 3

    params._print_as_xml()

    # params.a3 = 3

    # params.child0 = 3

    params._save_as_xml()

    params_loaded = ParamContainer(path_file=params._tag + ".xml")
    os.remove(params._tag + ".xml")

    assert params_loaded == params



"""
Test dantec module
====================

"""

import unittest
import os
from shutil import rmtree

from .. import dantec


class TestDantex(unittest.TestCase):
    """Test fluiddyn.io.dantec module."""

    def setUp(self):
        self._work_dir = "test_fluiddyn_io_dantec"
        if not os.path.exists(self._work_dir):
            os.mkdir(self._work_dir)

        os.chdir(self._work_dir)

    def tearDown(self):
        os.chdir("..")
        rmtree(self._work_dir)


if __name__ == "__main__":
    unittest.main()



"""
Test timer module
=================

"""

import unittest

from ..timer import Timer, TimerIrregular, time_gteq, TimeStr


class TestTimer(unittest.TestCase):
    """Test fluiddyn.util.timer module."""

    def test_timer(self):
        time_gteq("16:20:10", "17:20:10")

        timer = Timer(0.01)
        timer.wait_tick()

        timer = TimerIrregular([0, 0.01])
        timer.wait_tick()

    def test_timestr(self):
        assert TimeStr("1:20") == TimeStr("00:01:20")
        assert TimeStr("1:20") == "1-00:01:20"
        assert TimeStr("1:20") != "00:01:00"
        assert TimeStr("1:20") <= "00:01:20"
        assert TimeStr("1:20") < "2-00:01:20"
        assert TimeStr("1:20") >= "00:01:00"
        assert TimeStr("1:20") > "20"


if __name__ == "__main__":
    unittest.main()



"""
PBS clusters (:mod:`fluiddyn.clusters.pbs`)
===========================================

Provides:

.. autoclass:: ClusterPBS
   :members:

"""

from __future__ import print_function

from builtins import input
from builtins import range
import os

from . import subprocess
from .local import ClusterLocal


class ClusterPBS(ClusterLocal):
    """Base class for clusters with PBS job scheduler."""

    _doc_commands = """
Useful commands
---------------

qsub
qstat -u $USER
qdel
qhold
qrls"""
    name_cluster = ""
    nb_cores_per_node = 32
    default_project = None
    cmd_run = "mpirun"
    cmd_launch = "qsub"
    max_walltime = "23:59:59"

    def __init__(self):
        self.check_pbs()
        self.commands_setting_env = ["cd $PBS_O_WORKDIR"]
        self.commands_unsetting_env = []

    def check_pbs(self):
        """Check if this script is run on a frontal with pbs installed."""
        try:
            subprocess.check_call(["qsub", "--version"], stdout=subprocess.PIPE)
        except OSError:
            raise ValueError(
                "This script should be run on a cluster with PBS installed."
            )

    def check_name_cluster(self, env="HOSTNAME"):
        """Check if self.name_cluster matches the environment variable."""

        if self.name_cluster not in os.getenv(env):
            raise ValueError(
                "Cluster name mismatch detected; expected " + self.name_cluster
            )

    def submit_command(
        self,
        command,
        name_run="fluiddyn",
        nb_nodes=1,
        nb_cores_per_node=None,
        walltime=None,
        project=None,
        queue=None,
        nb_mpi_processes=None,
        omp_num_threads=1,
        nb_runs=1,
        path_launching_script=None,
        path_resume=None,
        retain_script=True,
        ask=True,
        bash=True,
        email=None,
        interactive=False,
        **kwargs
    ):
        """Submit a command.

        Parameters
        ----------
        command : string
            Command which executes the run
        name_run : string
            Name of the run to be displayed in PBS queue
        nb_nodes : integer
            Sets number of MPI processes = nb_nodes * nb_cores_per_node
        nb_cores_per_node : integer
            Defaults to a maximum is fixed for a cluster, as set by self.nb_cores_per_node.
            Set as 1 for a serial job. Set as 0 to spread jobs across nodes
            (starts job faster, maybe slower).
        walltime : string
            Minimum walltime for the job
        project : string
            Sets the allocation to run the job under
        queue: string
            Sets the cluster queue to run the job on
        nb_mpi_processes : integer
            Number of MPI processes. Defaults to a `nb_cores_per_node * nb_nodes`.
        omp_num_threads : integer
            Number of OpenMP threads
        nb_runs : integer
            Number of times to submit jobs (launch once using `command` and
            resume thereafter with `path_resume` script / command).
        path_launching_script: string
            Path of the PBS jobscript
        path_resume : string
            Path of the script to resume a job, which takes one
            argument - the `path_run` parsed from the output.
        retain_script : boolean
            Retail or delete script after launching job
        ask : boolean
            Ask for user input to submit the jobscript or not
        bash : boolean
            Submit jobscript via `call_bash` function
        email : string
            In case of failure notify to the specified email address
        interactive : boolean
            Use `cmd_run_interactive` instead of `cmd_run` inside the jobscript
        """
        nb_cores_per_node, nb_mpi_processes = self._parse_cores_procs(
            nb_nodes, nb_cores_per_node, nb_mpi_processes
        )

        path_launching_script = self._make_path_launching_script()

        self._check_walltime(walltime)

        is_resume_script = bool("resumer" in path_launching_script)
        if project is None:
            project = self.default_project

        launching_command = self.cmd_launch

        if is_resume_script:
            dependencies = input("Enter jobid dependencies :").split()
            launching_command += " -W depend=afternotok:" + ":".join(dependencies)
        else:
            dependencies = None

        create_txt_kwargs = locals()
        del create_txt_kwargs["self"]
        txt = self._create_txt_launching_script(**create_txt_kwargs)
        self._write_txt_launching_script(txt, path_launching_script)
        launching_command += " ./" + path_launching_script
        self._launch(launching_command, command, bash, ask)

        if not retain_script:
            os.remove(path_launching_script)

        nb_times_resume = int(nb_runs) - 1
        for n in range(0, nb_times_resume):
            nb_runs = 1
            path = path_resume
            path_launching_script = self._make_path_launching_script("resumer")
            submit_script_kwargs = locals()
            del submit_script_kwargs["self"]
            del submit_script_kwargs["command"]
            self.submit_script(**submit_script_kwargs)

    def _create_txt_launching_script(self, **kwargs):
        """
        Examples
        --------

            #!/bin/bash -l
            # The -l above is required to get the full environment with modules

            # The name of the script is myjob
            #PBS -N myjob

            # Set the allocation to be charged for this job
            # not required if you have set a default allocation
            #PBS -A 201X-X-XX

            # Set the queue to launch this job on
            #PBS -q queuename

            # Only 1 hour wall-clock time will be given to this job
            #PBS -l walltime=1:00:00

            # Number of nodes and processes per node
            #PBS -l nodes=1:ppn=32

            #PBS -e $PBS.myjob.$PBS_JOBID.stderr
            #PBS -o $PBS.myjob.$PBS_JOBID.stdout

            # Run the executable named myexe
            mpirun -n 128 ./myexe
        """
        path_launching_script = kwargs["path_launching_script"]
        command = kwargs["command"]
        name_run = kwargs["name_run"]
        project = kwargs["project"]
        queue = kwargs["queue"]
        nb_nodes = kwargs["nb_nodes"]
        nb_cores_per_node = kwargs["nb_cores_per_node"]
        walltime = kwargs["walltime"]
        nb_mpi_processes = kwargs["nb_mpi_processes"]
        omp_num_threads = kwargs["omp_num_threads"]
        dependencies = kwargs["dependencies"]
        email = kwargs["email"]
        is_resume_script = kwargs["is_resume_script"]

        logfile = f"PBS.{name_run}"
        logfile_stdout = logfile + ".${PBS_JOBID}.stdout"

        txt = "#!/bin/bash -l\n\n"

        txt += f"#PBS -N {name_run}\n\n"
        if project is not None:
            txt += f"#PBS -A {project}\n\n"

        if queue is not None:
            txt += f"#PBS -q {queue}\n\n"

        if walltime is not None:
            txt += f"#PBS -l walltime={walltime}\n"

        txt += f"#PBS -l nodes={nb_nodes}:ppn={nb_cores_per_node}\n"

        if email is not None:
            txt += "#PBS -m a\n"
            txt += f"#PBS -M {email}\n"

        txt += f"#PBS -e {logfile}.%J.stderr\n"
        txt += f"#PBS -o {logfile}.%J.stdout\n\n"

        txt += 'echo "hostname: "$HOSTNAME\n\n'
        txt += self._log_job(
            nb_mpi_processes,
            path_launching_script,
            logfile_stdout,
            command,
            "PBS_JOB.md",
        )

        txt += "\n".join(self.commands_setting_env) + "\n\n"

        if omp_num_threads is not None:
            txt += f"export OMP_NUM_THREADS={omp_num_threads}\n\n"

        if is_resume_script:
            jobid = dependencies[0]
            main_logfile = f"PBS.{name_run}.{jobid}.stdout"
            txt += "PATH_RUN=$(sed -n '/path_run/{n;p;q}' " + "{}\n".format(
                main_logfile
            )

        cmd = self.cmd_run

        if nb_mpi_processes > 1:
            txt += f"{cmd} -n {nb_mpi_processes} "

        if is_resume_script:
            txt += f"{command} $PATH_RUN"
        else:
            txt += command

        txt += "\n" + "\n".join(self.commands_unsetting_env)
        return txt



"""
LEGI clusters (:mod:`fluiddyn.clusters.legi`)
=============================================

Provides:

.. autoclass:: Calcul
   :members:

.. autoclass:: Calcul3
   :members:

.. autoclass:: Calcul9
   :members:

.. autoclass:: Calcul7
   :members:

"""
from __future__ import print_function
from subprocess import check_output
import time

from fluiddyn.clusters.oar import ClusterOAR


class Calcul3(ClusterOAR):
    name_cluster = "calcul3"
    has_to_add_name_cluster = True
    nb_cores_per_node = 12
    frontends = ["calcul8sv109", "legilnx44"]


class Calcul9(ClusterOAR):
    name_cluster = "calcul9"
    has_to_add_name_cluster = True
    nb_cores_per_node = 8
    frontends = ["calcul8sv109", "legilnx44"]


class Calcul(ClusterOAR):
    name_cluster = "calcul7-8"
    has_to_add_name_cluster = False
    nb_cores_per_node = 20
    frontends = ["nrj1sv223", "nrj1sv224"]
    job_count = 0

    def stall(self, name_run="fluiddyn", stall_after=1, t_sleep=30):
        """Stall job submission to wait for similar jobs to complete.

        Parameters
        ----------
        name_run: str
            Description of the job. Should be the same as in submit_script.
        stall_after: int
            Continue launching jobs, and stall when `job_count` is divisible
            by `stall_after`.
        t_sleep: int
            Time to sleep in seconds.

        """
        tstart = time.time()
        self.job_count += 1
        while (
            name_run in check_output(["oarstat", "-u"])
            and self.job_count % stall_after == 0
        ):
            tnow = time.time()
            print(
                "\rWaiting for",
                name_run,
                "to finish...",
                tnow - tstart,
                "seconds",
                end=" ",
            )
            time.sleep(t_sleep)


class Calcul7(Calcul):
    name_cluster = "calcul7"
    has_to_add_name_cluster = True
    nb_cores_per_node = 16


class Calcul8(Calcul):
    name_cluster = "calcul8"
    has_to_add_name_cluster = True



"""Minimal mpi (:mod:`fluiddyn.util.mpi`)
======================================

Minimal mpi module for sequential and simple mpi programs. It uses mpi4py if
installed but it can be imported in any cases, without ImportError.

Always defines the two variables ``nb_proc`` and ``rank`` and the function
``printby0``.

If the program as been launch using mpi, also defines the variable ``comm``
with is the ``MPI.COMM_WORLD`` communicator.

"""
from __future__ import print_function

import os
import sys
import traceback

import psutil


def _detect_mpi_type():
    """Detect MPI implementation using environent variable or process name.
    cf. http://stackoverflow.com/q/16612389

    """
    ENV_OVERRIDE = "FLUIDDYN_FORCE_MPI"
    MPIEXEC_TYPE = dict(
        {
            "PMI_RANK": "MPICH 2 and derivatives",
            "OMPI_COMM_WORLD_RANK": "OpenMPI >= 1.3",
            "OMPI_MCA_ns_nds_vpid": "OpenMPI 1.2 and derivatives",
            "PMI_ID": "SLURM PMI",
            "SLURM_PROCID": "SLURM",
            "SLURM_JOBID": "SLURM",
            "LAMRANK": "LAM",
            "MPI_RANKID": "HP MPI for Linux",
            "MP_CHILD": "IBM PE",
            "MP_RANK": "Sun CT",
            "MPIRUN_RANK": "MVAPICH >= 1.1",
            "MPI_FLAVOR": os.environ.get("MPI_FLAVOR"),
            ENV_OVERRIDE: "Unknown. Force loading mpi4py if available",
        }
    )

    set_mpiexec_type = set(MPIEXEC_TYPE.keys())
    set_os_env = set(os.environ.keys())
    set_detected_env = set_mpiexec_type.intersection(set_os_env)
    warning_msg = (
        "Warning: No known MPI environment variables detected.\n"
        "Try exporting environment variable {}=1 to skip MPI type check.\n"
    ).format(ENV_OVERRIDE)

    if hasattr(os, "getppid"):
        process_name = psutil.Process(os.getppid()).name()
    else:
        # Python 2.7 on Windows
        process_name = None

    if len(set_detected_env) > 0:
        env = set_detected_env.pop()
        if env == ENV_OVERRIDE and os.environ[env] != "1":
            print(warning_msg)
            return None, process_name
        else:
            return MPIEXEC_TYPE[env], process_name
    else:
        if any(process_name.startswith(name) for name in ("mpirun", "mpiexec")):
            print(
                warning_msg,
                "Loading MPI anyways, since the program was launched using ",
                process_name,
            )
            return MPIEXEC_TYPE[ENV_OVERRIDE], process_name

    # If none of the above cases match
    return None, process_name


class NoMPIError(Exception):
    pass


try:
    _mpi_type, process_name = _detect_mpi_type()
    if _mpi_type is None:
        raise NoMPIError
    elif _mpi_type == "MPT":
        # "arrayd" like "Array Services daemon"
        if process_name is None or process_name == "arrayd":
            from mpi4py import MPI
        else:
            raise NoMPIError
    else:
        from mpi4py import MPI
except (ImportError, NoMPIError):
    nb_proc = 1
    rank = 0
    printby0 = print
    print_sorted = print
else:
    comm = MPI.COMM_WORLD
    nb_proc = comm.size
    rank = comm.Get_rank()

    if nb_proc > 1:

        def my_excepthook(ex_cls, ex, tb):
            print("".join(traceback.format_exception(ex_cls, ex, tb)))
            sys.__excepthook__(ex_cls, ex, tb)
            MPI.COMM_WORLD.Abort()

        sys.excepthook = my_excepthook

    standard_print = print

    def printby0(*args, **kwargs):
        if rank == 0:
            standard_print(*args, **kwargs)

    def print_sorted(*args, **kwargs):
        kwargs["flush"] = True
        comm.barrier()
        for irank in range(nb_proc):
            if rank == irank:
                standard_print(f"rank {rank}:", flush=True)
                standard_print(*args, **kwargs)
            comm.barrier()



"""
Simple operations on signals (:mod:`fluiddyn.calcul.signal`)
============================================================

"""

from __future__ import division, print_function

from builtins import zip
from builtins import range
from builtins import object
import numpy as np
from scipy import ndimage
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt

from fluiddyn.output import show


def decimate(sig, q, nwindow=None, axis=-1):
    """Decimate a signal."""
    if nwindow is None:
        nwindow = q

    shape_decimated = list(sig.shape)
    len_axis = shape_decimated[axis]
    shape_decimated[axis] = int(len_axis / q)

    sigdec = np.empty(shape_decimated, dtype=sig.dtype)

    for inds, value in np.ndenumerate(sigdec):
        # print(inds)

        sl = list(inds)

        ind_axis = q * inds[axis]
        indmin_axis = max(ind_axis - nwindow, 0)
        indmax_axis = min(ind_axis + nwindow, len_axis - 1)

        sl[axis] = slice(indmin_axis, indmax_axis)

        sl = tuple(sl)

        # print(sl)
        # print(sig[sl])

        sigdec[inds] = sig[sl].mean()

    return sigdec


class FunctionLinInterp:
    """Function defined by a linear interpolation."""

    def __init__(self, x, f):
        if not isinstance(x, (list, tuple, np.ndarray)) or not isinstance(
            f, (list, tuple, np.ndarray)
        ):
            raise ValueError("x and f should be sequence.")

        self.x = np.array(x, np.float64)
        self.f = np.array(f, np.float64)

        # test for same length
        if len(x) != len(f):
            raise ValueError("len(x) != len(f)")

        # test for x increasing
        if any([x2 - x1 < 0 for x1, x2 in zip(self.x, self.x[1:])]):
            raise ValueError("x must be in ascending order!")

        self.func = interp1d(x, f)

    def __call__(self, x):
        return self.func(x)

    def plot(self):

        fig = plt.figure()

        size_axe = [0.13, 0.16, 0.84, 0.76]
        ax = fig.add_axes(size_axe)

        ax.set_xlabel(r"$x$")
        ax.set_ylabel(r"$y$")
        ax.plot(self.x, self.f, "k-.")

        show()


def deriv(f, x=None, dx=None, method="diff"):
    """Derive a 1D signal."""
    if dx is None:
        dx = np.diff(x)

    if method == "diff":
        x = (x[:-1] + x[1:]) / 2
        return x, np.diff(f) / dx

    elif method == "convolve":
        return np.convolve(f, [1, -1]) / dx

    elif method == "gaussian_filter":
        return ndimage.gaussian_filter1d(f, sigma=1, order=1, mode="wrap") / dx

    else:
        raise ValueError("unknown method...")


def smooth(x, window_len=11, window="hanning"):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with
    the signal.  The signal is prepared by introducing reflected
    copies of the signal (with the window size) in both ends so that
    transient parts are minimized in the begining and end part of the
    output signal.

    Parameters
    ----------

        x: the input signal
        window_len: the dimension of the smoothing window; should be
        an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming',
        'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    Returns
    -------

        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman,
    numpy.convolve, scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array
    instead of a string

    NOTE: length(output) != length(input), to correct this: return
    y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        ret = np.empty(x.shape)
        for i0 in range(x.shape[0]):
            ret[i0] = smooth(x[i0], window_len=window_len, window=window)
        return ret

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if window_len % 2 == 0:
        raise ValueError("window_len should be odd.")

    if window not in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        raise ValueError(
            "Window is on of 'flat', 'hanning', "
            "'hamming', 'bartlett', 'blackman'"
        )

    s = np.r_[x[window_len - 1 : 0 : -1], x, x[-1:-window_len:-1]]
    if window == "flat":  # moving average
        w = np.ones(window_len, "d")
    else:
        w = eval("np." + window + "(window_len)")

    y = np.convolve(w / w.sum(), s, mode="valid")
    return y[(window_len // 2 - 1) : -(window_len // 2 + 1)]


if __name__ == "__main__":

    # Data:
    x = np.linspace(0, 2 * np.pi, 100)
    f = np.sin(x) + 0.02 * (np.random.rand(100) - 0.5)

    # First derivatives:
    dx = x[1] - x[0]  # use np.diff(x) if x is not uniform
    df = np.diff(f) / dx
    cf = np.convolve(f, [1, -1]) / dx
    gf = ndimage.gaussian_filter1d(f, sigma=1, order=1, mode="wrap") / dx

    # Second derivatives:
    dxdx = dx ** 2
    ddf = np.diff(f, 2) / dxdx
    ccf = np.convolve(f, [1, -2, 1]) / dxdx
    ggf = ndimage.gaussian_filter1d(f, sigma=1, order=2, mode="wrap") / dxdx

    sig = np.zeros([4, 4, 4])

    sigd = decimate(sig, 2, nwindow=2, axis=0)



"""
Scientific output toolkit
=========================

.. _output:
.. currentmodule:: fluiddyn.output

.. autosummary::
   :toctree:

   figs
   rcparams
   util
   colorchart

"""

# temporary to avoid the error where matplotlib is not installed
try:
    from fluiddyn.output.figs import show
    from fluiddyn.output.rcparams import set_rcparams
    from fluiddyn.output.util import gradient_colors
    import fluiddyn

    fluiddyn.show = show
except ImportError:
    pass

__all__ = ["show", "set_rcparams", "gradient_colors"]



"""Matlab -> wrong Python translator
====================================

Many more simple transformations can be implemented, for example:

+----------------+---------------------+
|   ``disp(``    |    ``print(``       |
+----------------+---------------------+
| ``zeros(...)`` | ``np.zeros((...))`` |
+----------------+---------------------+

"""

from __future__ import print_function

import os

from . import cleanmat
from .cleanmat import modif_code

block_definers = ("if", "while", "for", "elseif")

ident = 4 * " "


def is_comment_line(line):
    return line.strip().startswith("#")


def modif_blocks(code_lines):
    lines_new = []
    for line in code_lines:

        if not any(line.strip().startswith(s) for s in block_definers + ("end",)):
            lines_new.append(line)
            continue

        if not line.strip().startswith("end"):
            lines_new.append(line + ":")

    return lines_new


def modif_comments(code_lines):
    lines_new = []

    for line in code_lines:
        if not cleanmat.is_comment_line(line):
            lines_new.append(line)
            continue

        lines_new.append(line.replace("%", "#", 1))

    return lines_new


def modif_remove_semicolon(code_lines):
    lines_new = []

    for line in code_lines:
        if is_comment_line(line):
            lines_new.append(line)
            continue

        if line.endswith(";"):
            line = line[:-1]

        lines_new.append(line)

    return lines_new


def modif_remove_3dots(code_lines):
    lines_new = []

    for line in code_lines:
        if is_comment_line(line):
            lines_new.append(line)
            continue

        if line.endswith("..."):
            line = line[:-3] + "\\"

        lines_new.append(line)

    return lines_new


def create_py_code(path_file):

    code = modif_code(path_file)

    code_lines = code.split("\n")

    modif_functions = [
        modif_blocks,
        modif_comments,
        modif_remove_semicolon,
        modif_remove_3dots,
    ]

    for func in modif_functions:
        code_lines = func(code_lines)

    new_code = "\n".join(code_lines)
    return new_code


def treat_matlab_directory(path_dir):

    path_cleaner = path_dir + "_py"

    if not os.path.exists(path_cleaner):
        os.mkdir(path_cleaner)

    for root, subdirs, nfiles in os.walk(path_dir):
        relative_path = root[len(path_dir) :]

        new_dir = path_cleaner + relative_path

        if not os.path.exists(new_dir):
            os.mkdir(new_dir)

        matlab_nfiles = [nfile for nfile in nfiles if nfile.endswith(".m")]

        for nfile in matlab_nfiles:
            nfilepy = nfile[:-2] + ".py"

            path_file = os.path.join(root, nfile)
            path_file_new = os.path.join(new_dir, nfilepy)

            print((path_file, path_file_new))

            new_code = create_py_code(path_file)

            with open(path_file_new, "w") as f:
                f.write(new_code)


if __name__ == "__main__":

    path_dir = "diablo_mat"

    treat_matlab_directory(path_dir)

# code = modif_code(path_dir + '/diablo.m')
# print(code)

# code_lines = [
#     "    TIME=TIME+DELTA_T;",
#     "    if (mod(TIME_STEP,10)==0 & toto == 2)"]
# print('\n'.join(modif_spaces_around_operators(code_lines)))



import unittest
import os
import sys

from shutil import rmtree
from runpy import run_path

import numpy as np

from ..in_py import save_in_py


class TestInPy(unittest.TestCase):
    """Test fluiddyn.io.in_py module."""

    @classmethod
    def setUpClass(cls):
        cls._work_dir = "test_fluiddyn_io_in_py"
        if not os.path.exists(cls._work_dir):
            os.mkdir(cls._work_dir)

        os.chdir(cls._work_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test(self):
        a = 1
        b = np.ones(2)
        c = np.ones([2, 2])
        d = 2
        save_in_py("file_globals.py", locals(), ("a", "b", "c"))

        # for python 3.7
        sys.path.insert(0, "")

        import file_globals as fg

        self.assertEqual(a, fg.a)
        self.assertTrue(np.allclose(b, fg.b))
        self.assertTrue(np.allclose(c, fg.c))

        d = {"a": a, "b": b, "c": d}
        save_in_py("file_dict.py", d)

        fd = run_path("file_dict.py")
        self.assertEqual(d["a"], fd["a"])
        self.assertTrue(np.allclose(d["b"], fd["b"]))
        self.assertTrue(np.allclose(d["c"], fd["c"]))


if __name__ == "__main__":
    unittest.main()



"""Utilities to read RD Vision files
====================================

"""

import os
import configparser

try:
    from PIL import Image
except ImportError:
    pass

import numpy as np

from fluiddyn.util.paramcontainer import ParamContainer, tidy_container
from fluiddyn.io.binary import BinFile


def read_seq(name):
    if not name.endswith(".seq"):
        name += ".seq"
    config = configparser.ConfigParser()
    config.read([name])
    l = config.items(config.sections()[0])
    d = {k.replace(" ", "_"): v for k, v in l}
    return d


def read_xml(name):
    if not name.endswith(".xml"):
        name += ".xml"

    p = ParamContainer(path_file=name)
    tidy_container(p)
    return p


def read_sqb(name, nb_files=1):
    if not name.endswith(".sqb"):
        name += ".sqb"
    offsets = np.empty([nb_files], dtype=np.uint32)
    timestamps = np.empty([nb_files], dtype=np.float64)
    indices_files = np.empty([nb_files], dtype=np.uint32)

    with BinFile(name) as f:
        for i in range(nb_files):
            offsets[i] = f.readt(1, "uint32")
            f.readt(1, "uint32")
            timestamps[i] = f.readt(1, "float64")
            indices_files[i] = f.readt(1, "uint32")
            f.readt(1, "uint32")

    return offsets, timestamps, indices_files


class SetOfFiles:
    def __init__(self, name):

        self.d = d = read_seq(name)

        self.name = d["sequence_name"]
        self.name_short = name.split("_")[0]

        self.width = int(d["width"])
        self.height = int(d["height"])
        self.bytesperpixel = int(d["bytesperpixel"])
        self.bitsperpixel = int(d["bitsperpixel"])

        self.nb_files = int(d["number_of_files"])

        self.offsets, self.timestamps, self.indices_files = read_sqb(
            name + ".sqb", int(d["number_of_files"])
        )

    def read_im(self, index):
        path = (
            self.d["bin_repertoire"]
            + "/"
            + self.d["bin_file"]
            + "{:05d}.bin".format(self.indices_files[index])
        )

        assert os.path.exists(path)

        with BinFile(path) as f:
            f.seek(self.offsets[index])
            im = np.reshape(
                f.readt(
                    self.width * self.height,
                    "uint{}".format(self.bytesperpixel * 8),
                ),
                [self.height, self.width],
            ).astype(np.dtype("int32"))

        return im * 2 ** (16 - self.bitsperpixel)

    def convert_all_images(self):
        new_dir = self.name_short
        if not os.path.exists(new_dir):
            os.makedirs(new_dir)

        for index in range(self.nb_files):
            im = self.read_im(index)
            path = os.path.join(new_dir, f"im_{index}.png")
            im_pil = Image.fromarray(im)
            im_pil.save(path)



from __future__ import print_function

import os
import unittest
import smtplib

import matplotlib.pyplot as plt

from ..logger import Logger


class UnitTestException(Exception):
    pass


def mySMTP(*args, **kwargs):
    raise UnitTestException


# monkey patch
smtplib.SMTP = mySMTP


class TestFFTW1DReal2Complex(unittest.TestCase):
    def tearDown(self):
        os.remove("storage_file.txt")

    def test_logger(self):
        logger = Logger(
            "storage_file.txt",
            "test@test.org",
            "test@test.com",
            email_title="test",
            email_delay=0.02,
        )
        logger.print_log("")

        with self.assertRaises(UnitTestException):
            logger.send_email_if_has_to(name_exception="", figures=[plt.figure()])

        with self.assertRaises(UnitTestException):
            raise UnitTestException


if __name__ == "__main__":
    unittest.main()



"""Console script to print and save system information. (:mod:`fluiddyn.util.info`)
===================================================================================
Displays all important information related to software and hardware. It also
includes detailed information such as currently installed FluidDyn packages,
other third-party packages, C compiler, MPI and NumPy configuration.

Examples
--------
>>> fluidinfo  # print package, Python, software and hardware info
>>> fluidinfo -v  # also print basic Numpy info
>>> fluidinfo -vv  # also print detailed Numpy info
>>> fluidinfo -s  # save all information into sys_info.xml
>>> fluidinfo -o /tmp  # save all information into /tmp/sys_info.xml


.. todo::
    Use a YAML package to print.

"""
from __future__ import print_function
from importlib import import_module as _import
import os
import shlex
import inspect
from collections import OrderedDict
import platform
import argparse
import warnings
import shutil
from pathlib import Path
from configparser import ConfigParser

try:
    from distro import linux_distribution
except ImportError:
    try:
        from platform import linux_distribution
    # Pending deprecation (Python 3.7)
    except ImportError:
        ImportError("Install distro package to use this module.")

import subprocess

import psutil
import numpy as np
import numpy.distutils.system_info as np_sys_info
import numpy.__config__ as np_build_info

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=ImportWarning)
    from ..io.redirect_stdout import stdout_redirected
    from .paramcontainer import ParamContainer
    from .terminal_colors import cprint

_COL_WIDTH = 32
_COL_MAX = 80


def safe_check_output(cmd, first_row_only=True):
    """Error-tolerant version of subprocess check output"""
    cmd = f'/bin/sh -c "{cmd}; exit 0"'
    output = subprocess.check_output(
        shlex.split(cmd), stderr=subprocess.STDOUT
    ).decode("utf-8")

    if first_row_only and output != "":
        return output.splitlines()[0]

    else:
        return output


def _get_hg_repo(path_dir):
    """Parse `hg paths` command to find remote path."""
    if path_dir == "":
        return ""

    hgrc = Path(path_dir) / ".hg" / "hgrc"
    if hgrc.exists():
        config = ConfigParser()
        config.read(str(hgrc))
        if "paths" in config:
            return config["paths"].get("default", "hgrc: no default path?")
        else:
            return "hgrc: no [paths] section?"
    else:
        return "not a hg repo"


def make_dict_about(pkg):
    """Make dictionary with all collected information about one package."""
    about_pkg = OrderedDict(
        [
            ("installed", None),
            ("version", ""),
            ("local_path", ""),
            ("remote_path", ""),
        ]
    )
    try:
        pkg = _import(pkg)
    except ImportError:
        about_pkg["installed"] = False
        return about_pkg

    else:
        about_pkg["installed"] = True
        about_pkg["version"] = pkg.__version__
        init_file = inspect.getfile(pkg)
        if "site-packages" in init_file or "dist-packages" in init_file:
            about_pkg["local_path"] = os.path.dirname(init_file)
            about_pkg["remote_path"] = ""
        else:
            about_pkg["local_path"] = os.path.dirname(os.path.dirname(init_file))
            about_pkg["remote_path"] = _get_hg_repo(about_pkg["local_path"])
        return about_pkg


def get_info_python():
    """Python information."""
    info_py = OrderedDict.fromkeys(["version", "implementation", "compiler"])
    for k in info_py:
        func = getattr(platform, "python_" + k)
        info_py[k] = func()

    return info_py


def _get_info(pkgs):
    """Create a dictionary of dictionaries for all packages."""
    for pkg in pkgs:
        dict_pkg = make_dict_about(pkg)
        pkgs[pkg] = dict_pkg

    return pkgs


def get_info_fluiddyn():
    """Create a dictionary of dictionaries for all FluidDyn packages."""
    pkgs = OrderedDict.fromkeys(
        [
            "fluiddyn",
            "fluidsim",
            "fluidlab",
            "fluidimage",
            "fluidfft",
            "fluidsht",
            "fluidcoriolis",
            "fluiddevops",
            "transonic",
        ]
    )
    return _get_info(pkgs)


def get_info_third_party():
    """Create a dictionary of dictionaries for all third party packages."""
    pkgs = OrderedDict.fromkeys(
        [
            "numpy",
            "cython",
            "mpi4py",
            "pythran",
            "pyfftw",
            "matplotlib",
            "scipy",
            "skimage",
            "h5py",
        ]
    )
    return _get_info(pkgs)


def get_info_software():
    """Create a dictionary for compiler and OS information."""
    uname = platform.uname()
    info_sw = OrderedDict(zip(["system", "hostname", "kernel"], uname))
    try:
        info_sw["distro"] = " ".join(linux_distribution())
    except Exception:
        pass

    cc = os.getenv("CC", "gcc")

    info_sw["CC"] = safe_check_output(cc + " --version")
    info_sw["MPI"] = safe_check_output("mpirun --version")
    return info_sw


def get_info_numpy(only_print=False, verbosity=None):
    """Print or create a dictionary for numpy and linalg library information."""
    libs = ["lapack_opt", "blas_opt"]
    libs_sys = ["fftw"]

    def rm_configtest():
        if os.path.exists("_configtest.o.d"):
            os.remove("_configtest.o.d")

    def np_get_info():
        with stdout_redirected():
            np_sys_info_dict = OrderedDict(
                (k, np_sys_info.get_info(k)) for k in libs + libs_sys
            )
            np_build_info_dict = OrderedDict(
                (k, np_build_info.get_info(k)) for k in libs
            )

        rm_configtest()
        return np_sys_info_dict, np_build_info_dict

    if only_print:
        if verbosity == 1:
            np_sys_info_dict, np_build_info_dict = np_get_info()

            _print_dict(np_sys_info_dict, subheading="system")
            _print_dict(np_build_info_dict, subheading="\nbuild")
        elif verbosity == 2:
            np_sys_info.show_all(argv=[])
            np.show_config()
            rm_configtest()
    else:
        np_sys_info_dict, np_build_info_dict = np_get_info()
        return np_sys_info_dict, np_build_info_dict


def get_info_h5py():
    """Create a dictionary detailing h5py installation."""
    try:
        import h5py
    except ImportError:
        return {}

    config = h5py.get_config()
    hdf5_version = h5py.version.hdf5_version_tuple
    try:
        vds = hdf5_version > config.vds_min_hdf5_version
    except AttributeError:
        vds = False

    try:
        swmr = hdf5_version > config.swmr_min_hdf5_version
    except AttributeError:
        swmr = False

    info = OrderedDict(
        (
            ("HDF5_version", h5py.version.hdf5_version),
            ("MPI_enabled", config.mpi),
            ("virtual_dataset_available", vds),
            ("single_writer_multiple_reader_available", swmr),
        )
    )
    return info


def filter_modify_dict(d, filter_keys, mod_keys):
    """Create a new dictionary by filtering and modifying the keys."""
    filter_d = OrderedDict((k, v) for k, v in d.items() if k in filter_keys)
    for old_key, new_key in zip(filter_keys, mod_keys):
        if old_key != new_key:
            try:
                filter_d[new_key] = filter_d.pop(old_key)
            except KeyError:
                pass

    return filter_d


def update_dict(d1, d2):
    """Update dictionary with missing keys and related values."""
    d1.update(((k, v) for k, v in d2.items() if k not in d1))
    return d1


def get_info_hardware():
    """Create a dictionary for CPU information."""

    def _cpu_freq():
        """psutil can return `None` sometimes, esp. in Travis."""
        func = "psutil.cpu_freq: "
        try:
            hz = psutil.cpu_freq()
        except IOError:
            return (func + "IOError",) * 3  # See psutil issue #1071

        except AttributeError:
            return (func + "AttributeError",) * 3  # See psutil issue #1006

        except NotImplementedError:
            return (
                func + "NotImplementedError",
            ) * 3  # on occigen (cluster cines)

        if hz is None:
            return (func + "None",) * 3  # See psutil issue #981

        else:
            return (f"{h:.3f}" for h in hz)

    try:
        from numpy.distutils.cpuinfo import cpu

        # Keys are specific to Linux distributions only
        info_hw = filter_modify_dict(
            cpu.info[0],
            [
                "uname_m",
                "address sizes",
                "bogomips",
                "cache size",
                "model name",
                "cpu cores",
                "siblings",
            ],
            [
                "arch",
                "address_sizes",
                "bogomips",
                "cache_size",
                "cpu_name",
                "nb_cores",
                "nb_siblings",
            ],
        )
        info_hw["cpu_MHz_actual"] = []
        for d in cpu.info:
            info_hw["cpu_MHz_actual"].append(float(d["cpu MHz"]))
    except KeyError as e:
        print("KeyError with", e)
        info_hw = OrderedDict()

    hz_current, hz_min, hz_max = _cpu_freq()
    info_hw_alt = OrderedDict(
        (
            ("arch", platform.machine()),
            ("cpu_name", platform.processor()),
            ("nb_procs", psutil.cpu_count()),
            ("cpu_MHz_current", hz_current),
            ("cpu_MHz_min", hz_min),
            ("cpu_MHz_max", hz_max),
        )
    )
    info_hw = update_dict(info_hw, info_hw_alt)
    return info_hw


def reset_col_width(widths):
    """Detect total width of the current terminal."""
    global _COL_MAX, _COL_WIDTH
    nb_cols = len(widths)
    _COL_MAX = shutil.get_terminal_size().columns
    _COL_WIDTH = _COL_MAX // nb_cols

    widths_array = np.array(widths)
    widths_sum = widths_array.sum()
    if widths_sum > _COL_MAX:
        too_long = np.greater(widths_array, _COL_WIDTH, dtype=int)
        nb_too_long = np.count_nonzero(too_long)
        widths_too_long = widths_array[too_long].sum()

        _COL_WIDTH = (_COL_MAX - (widths_sum - widths_too_long)) // nb_too_long
        widths = [_COL_WIDTH if w > _COL_WIDTH else w for w in widths]

    return widths


# Table formatting functions


def _print_item(item, color=None, bold=False, width=None):
    if width is None:
        width = _COL_WIDTH

    cprint(item.ljust(width), end="", color=color, bold=bold)


def _print_heading(heading, underline_with="=", case="title", widths=None):
    if isinstance(heading, str):
        heading = [heading]

    if widths is None:
        widths = [None] * len(heading)

    if case == "title":
        heading = [h.replace("_", " ").title() for h in heading]
    elif case == "upper":
        heading = [h.replace("_", " ").upper() for h in heading]

    underline = [underline_with * len(h) for h in heading]

    for item, w in zip(heading, widths):
        _print_item(item, color="RED", bold=True, width=w)

    print()
    for item, w in zip(underline, widths):
        _print_item(item, color="RED", width=w)

    print()


def _print_dict(
    d,
    heading=None,
    underline_with="=",
    case="title",
    subheading=None,
    indent_level=0,
):
    if heading is not None:
        _print_heading("\n" + heading, underline_with, case)

    indent = " " * indent_level * 2
    WIDTH = _COL_WIDTH - indent_level * 2

    if subheading is not None:
        cprint(f"{indent}{subheading}:", color="BLUE")

    for k, v in d.items():
        if isinstance(v, dict):
            _print_dict(v, subheading=k, indent_level=indent_level + 1)
        else:
            print("{}  - {}: {}".format(indent, k.ljust(WIDTH), v))


def get_col_widths(d, widths=None):
    if widths is None:
        widths = [len(key) + 1 for key in d]

    for i, string in enumerate(d.values()):
        widths[i] = max(widths[i], len(repr(string)) + 1)

    return widths


def print_sys_info(verbosity=None):
    """Print package information as a formatted table."""
    global _COL_MAX
    pkgs = get_info_fluiddyn()
    pkgs_third_party = get_info_third_party()

    widths = None
    width_pkg_name = 0
    for _pkgs in (pkgs, pkgs_third_party):
        for pkg_name, pkg_details in _pkgs.items():
            width_pkg_name = max(width_pkg_name, len(pkg_name) + 1)
            widths = get_col_widths(pkg_details, widths)

    widths.insert(0, width_pkg_name)

    pkgs_keys = list(pkgs)
    heading = ["Package"]
    heading.extend(pkgs[pkgs_keys[0]])
    widths = reset_col_width(widths)

    _print_heading(heading, widths=widths)

    def print_pkg(about_pkg):
        for v, w in zip(about_pkg.values(), widths[1:]):
            v = str(v)
            if len(v) > w:
                v = v[:10] + "..." + v[10 + 4 - w :]
            _print_item(str(v), width=w)

        print()

    for pkg_name, about_pkg in pkgs.items():
        print(pkg_name.ljust(widths[0]), end="")
        print_pkg(about_pkg)

    for pkg_name, about_pkg in pkgs_third_party.items():
        print(pkg_name.ljust(widths[0]), end="")
        print_pkg(about_pkg)

    info_sw = get_info_software()
    _print_dict(info_sw, "Software")
    info_hw = get_info_hardware()
    _print_dict(info_hw, "Hardware")
    info_py = get_info_python()
    _print_dict(info_py, "Python")
    if verbosity is not None:
        _print_heading("\nNumPy", case=None)
        get_info_numpy(True, verbosity)
        info_h5py = get_info_h5py()
        _print_dict(info_h5py, "h5py", case=None)

    print()


def save_sys_info(path_dir=".", filename="sys_info.xml"):
    """Save all system information as a xml file."""

    sys_info = ParamContainer("sys_info")
    info_sw = get_info_software()
    info_hw = get_info_hardware()
    info_py = get_info_python()
    info_np = get_info_numpy()
    info_h5py = get_info_h5py()
    pkgs = get_info_fluiddyn()
    pkgs_third_party = get_info_third_party()

    sys_info._set_child("software", info_sw)
    sys_info._set_child("hardware", info_hw)
    sys_info._set_child("python", info_py)
    for pkg in pkgs:
        sys_info.python._set_child(pkg, pkgs[pkg])

    for pkg in pkgs_third_party:
        sys_info.python._set_child(pkg, pkgs_third_party[pkg])

    sys_info.python.numpy._set_child("system")
    sys_info.python.numpy._set_child("build")
    for k, v in info_np[0].items():
        sys_info.python.numpy.system._set_child(k, v)
    for k, v in info_np[1].items():
        sys_info.python.numpy.build._set_child(k, v)

    sys_info.python.h5py._set_child("config", info_h5py)
    path = os.path.join(path_dir, filename)
    sys_info._save_as_xml(path, find_new_name=True)


def _get_parser():
    desc = "\n".join(__doc__.splitlines()[2:])
    parser = argparse.ArgumentParser(
        prog="fluidinfo",
        description=desc,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.set_defaults(func=print_sys_info)
    parser.add_argument(
        "-s",
        "--save",
        help="saves system information to an xml file " "(sys_info.xml)",
        action="store_true",
    )
    parser.add_argument(
        "-o", "--output-dir", help="save to directory", default=None
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        help="increase output verbosity (max: -vvv)",
        action="count",
    )
    parser.add_argument(
        "-W", "--warnings", help="show warnings", action="store_true"
    )
    return parser


def main(args=None):
    """Parse arguments and execute ``fluidinfo``."""
    if args is None:
        parser = _get_parser()
        args = parser.parse_args()

    if not args.warnings:
        warnings.filterwarnings("ignore", category=UserWarning)

    if args.save:
        save_sys_info()
    elif args.output_dir is not None:
        save_sys_info(args.output_dir)
    else:
        print_sys_info(args.verbosity)

    if not args.warnings:
        warnings.resetwarnings()


if __name__ == "__main__":
    main()



"""
Test hdf5 module
================

"""

import unittest
import os
from shutil import rmtree

import numpy as np

from ..hdf5 import H5File, save_variables_h5, load_variables_h5


class TestHdf5(unittest.TestCase):
    """Test fluiddyn.io.hdf5 module."""

    @classmethod
    def setUpClass(cls):
        cls._work_dir = "test_fluiddyn_io_hdf5"
        if not os.path.exists(cls._work_dir):
            os.mkdir(cls._work_dir)

        os.chdir(cls._work_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test_dict(self):
        path = "test_dict.h5"
        d = {"a": 1, "b": 2}

        with H5File(path, "w") as f:
            f.save_dict("d", d)

        with H5File(path) as f:
            d1 = f.load_dict("d")

        self.assertEqual(d, d1)

    def test_dict_of_ndarrays(self):
        name = "myfile1.h5"
        d = {"times": 0.0, "a": np.ones(10), "b": np.zeros([2, 2])}

        with H5File(name, "w") as f:
            f.save_dict_of_ndarrays(d)

        with H5File(name) as f:
            d1 = f.load()

            for k, v in d.items():
                v1 = d1[k]

                self.assertTrue(np.allclose(v, v1))

        with H5File(name, "a") as f:
            d["times"] = 1.0
            f.save_dict_of_ndarrays(d)

            f.load(times_slice=[0, 2, 0.1])

    def test_functions(self):
        path = "test_functions0.h5"

        a = 1
        b = "str"
        c = np.ones(2)
        d = 10

        save_variables_h5(path, locals(), ("a", "b", "c"))
        variables1 = load_variables_h5(path)

        self.assertEqual(a, variables1["a"])
        self.assertEqual(b, variables1["b"])
        self.assertTrue(np.allclose(c, variables1["c"]))

        path = "test_functions1.h5"
        variables = {"a": a, "b": b, "c": c}
        save_variables_h5(path, variables)
        variables2 = load_variables_h5(path)

        self.assertEqual(a, variables2["a"])
        self.assertEqual(b, variables2["b"])
        self.assertTrue(np.allclose(c, variables2["c"]))


if __name__ == "__main__":
    unittest.main()



"""Input/output streams
=======================

.. _io:

Loading data from files and saving data to files are very common
tasks. However, we can loose a lot of time with silly problems. This subpackage
provides utilities for input/output to different file formats:

.. autosummary::
   :toctree:

   binary
   txt
   mycsv
   hdf5
   digiflow
   dantec
   davis
   rdvision
   ns3d
   multitiff
   in_py
   image
   redirect_stdout
   dump
   tee
   query


To define environment variables, use something like (on POSIX systems, in the
terminal or in your .bashrc)::

  export FLUIDSIM_PATH="/fsnet/project/meige/2015/15DELDUCA/DataSim"

"""

from builtins import str
import os
import pathlib

from .redirect_stdout import stdout_redirected

Path = pathlib.Path

HOME_DIR = Path.home()

FLUIDDYN_PATH_EXP = os.environ.get("FLUIDDYN_PATH_EXP")
if FLUIDDYN_PATH_EXP is not None:
    raise DeprecationWarning(
        "FLUIDDYN_PATH_EXP is deprecated: use FLUIDLAB_PATH."
    )

del FLUIDDYN_PATH_EXP

FLUIDLAB_PATH = os.environ.get("FLUIDLAB_PATH")
if FLUIDLAB_PATH is None:
    FLUIDLAB_PATH = str(HOME_DIR / "Exp_data")


FLUIDDYN_PATH_SIM = os.environ.get("FLUIDDYN_PATH_SIM")
if FLUIDDYN_PATH_SIM is not None:
    raise DeprecationWarning(
        "FLUIDDYN_PATH_SIM is deprecated: use FLUIDSIM_PATH."
    )

del FLUIDDYN_PATH_SIM

FLUIDSIM_PATH = os.environ.get("FLUIDSIM_PATH")
if FLUIDSIM_PATH is None:
    FLUIDSIM_PATH = str(HOME_DIR / "Sim_data")


FLUIDDYN_PATH_SCRATCH = os.getenv("FLUIDDYN_PATH_SCRATCH", FLUIDSIM_PATH)

FLUIDDYN_PATH_WARNING = os.environ.get("FLUIDDYN_PATH_WARNING")
if FLUIDDYN_PATH_WARNING is None:
    FLUIDDYN_PATH_WARNING = str(HOME_DIR / ".fluiddyn")


if not os.path.exists(FLUIDDYN_PATH_WARNING):
    os.makedirs(FLUIDDYN_PATH_WARNING)


def _write_warning(*args, **kargs):
    if "end" in kargs:
        end = kargs["end"]
    else:
        end = "\n"
    strings = []
    for i, v in enumerate(args):
        strings.append(str(v))
    with open(os.path.join(FLUIDDYN_PATH_WARNING, "warnings.txt"), "a") as f:
        f.write(" ".join(strings) + end)


__all__ = ["stdout_redirected"]



import unittest
from numpy.testing import assert_allclose

try:
    import shtns
except ImportError:
    use_shtns = False
else:
    use_shtns = True

from ...io import stdout_redirected
from .. import sphericalharmo


class TestSphericalHarmo(unittest.TestCase):
    """Test ``sphericalharmo`` module."""

    @unittest.skipUnless(use_shtns, "SHTns not installed or can not be imported.")
    def test_sht_random(self):
        """Test forward and inverse SHT on a random array."""
        with stdout_redirected():
            op = sphericalharmo.EasySHT(lmax=15)

        field_phys = op.create_array_spat("rand")
        field_spect = op.sht(field_phys)
        field_phys = op.isht(field_spect)

        field_spect = op.sht(field_phys)
        field_phys2 = op.isht(field_spect)

        assert_allclose(field_phys, field_phys2)


if __name__ == "__main__":
    unittest.main()



"""
Toolkit (:mod:`fluiddyn.util.util`)
===================================

"""

from __future__ import division, print_function

from builtins import object, str
import os
import sys
import glob
import inspect
import shutil
import datetime
from importlib import import_module
import contextlib

import psutil
import numpy as np

from ..io import Path, HOME_DIR
from . import mpi


def modification_date(path):
    """Modification date of a file or directory"""
    t = os.path.getmtime(path)
    return datetime.datetime.fromtimestamp(t)


def import_class(module_name, class_name):
    """Import a class from its name."""
    try:
        module = import_module(module_name)
    except ImportError as err:
        msg = "Error while attempting to import module " + module_name + "\n"
        raise ImportError(msg + err.args[0])

    try:
        return module.__dict__[class_name]

    except KeyError:
        raise KeyError(
            'The given class_name "' + class_name + '" is wrong. '
            "module_name: " + module_name
        )


def ipydebug():
    """Launch an Ipython shell"""

    # it is not useful to import this if the function is not used
    from IPython.terminal.embed import InteractiveShellEmbed

    ipshell = InteractiveShellEmbed()

    frame = inspect.currentframe().f_back
    msg = "Stopped at {0.f_code.co_filename} at line {0.f_lineno}".format(frame)

    # Go back one level!  This is needed because the call to
    # ipshell is inside this function.
    ipshell(msg, stack_depth=2)


def time_as_str(decimal=0):
    """Return a string coding the time."""
    dt = datetime.datetime.now()
    ret = dt.strftime("%Y-%m-%d_%H-%M-%S")
    if decimal > 0:
        if not isinstance(decimal, int):
            raise TypeError

        ret += f".{dt.microsecond:06d}"[: decimal + 1]
    return ret


def copy_me_in(dest=HOME_DIR):
    """Copy the file from where this function is called."""
    stack = inspect.stack()
    # bug Python 2 when using os.chdir
    path_caller = Path(stack[1][1])
    name_file = path_caller.name
    path_dest = (
        (Path(dest) / (name_file + "_" + time_as_str())).expanduser().absolute()
    )
    shutil.copyfile(str(path_caller), str(path_dest))
    return path_dest


def get_pathfile_from_strpath(str_path, ext="h5"):
    """Get the path of a file with a particular extension from a string."""
    if str_path.endswith("." + ext) and os.path.exists(str_path):
        return str_path

    if os.path.isdir(str_path):
        path_glob = str_path + os.path.sep + "*"
    else:
        path_glob = "*" + str_path + "*"

    paths = glob.glob(path_glob)
    for p in paths:
        if p.endswith("." + ext):
            return p

    raise ValueError(
        "Haven't been able to find a path corresponding to str_path."
        "(str_path: {}).".format(str_path)
    )


def create_object_from_file(str_path, *args, **kwargs):
    """Create an object from a file."""
    import h5py

    path = get_pathfile_from_strpath(str_path, ext="h5")

    # temporary... for compatibility
    try:
        with h5py.File(path, "r+") as f:
            keys = list(f.attrs.keys())
            if "class" in keys and "class_name" not in keys:
                f.attrs["class_name"] = f.attrs["class"]
            if "module_tank" in keys and "module_name" not in keys:
                f.attrs["module_name"] = f.attrs["module_tank"]
            else:
                if path.endswith("tank.h5") and "module_name" not in keys:
                    f.attrs["module_name"] = "fluiddyn.lab.tanks"
    except IOError:
        pass

    with h5py.File(path, "r") as f:
        class_name = f.attrs["class_name"]
        module_name = f.attrs["module_name"]

    if isinstance(class_name, np.ndarray):
        class_name = class_name[0]
        module_name = module_name[0]

    if not isinstance(class_name, str):
        class_name = class_name.decode()

    if not isinstance(module_name, str):
        module_name = module_name.decode()

    # # temporary... for compatibility
    # if class_name.startswith('<class '):
    #     class_name = class_name[8:-2].split('.')[-1]
    #     with h5py.File(path, 'r+') as f:
    #         f.attrs['class_name'] = class_name

    # module_name = fld._verif_names_modules(
    #     module_name, path, key_file='module_name')

    # fromlist has to be a not-empty so that __import__('A.B',
    # ...)  returns B rather than A.
    # module_exp = __import__(module_exp, fromlist=['not empty str'])

    module = import_module(module_name)  # .decode("utf-8"))
    Class = module.__dict__[class_name]  # .decode("utf-8")]

    return Class(*args, str_path=str_path, **kwargs)


def run_from_ipython():
    raise DeprecationWarning(
        "run_from_ipython() is deprecated. Use is_run_from_ipython instead."
    )

    is_run_from_ipython()


def is_run_from_ipython():
    """Check whether the code is run from Ipython."""
    try:
        __IPYTHON__
        return True

    except NameError:
        return False


def is_run_from_jupyter():
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
    except NameError or AttributeError:
        return False

    else:
        if shell == "ZMQInteractiveShell":
            return True

        else:
            return False


class Params:
    """Minimalist object to store some parameters."""

    def __repr__(self):
        return dict.__repr__(self.__dict__)


def get_memory_usage():
    """Return the memory usage in Mo."""
    process = psutil.Process(os.getpid())
    mem = process.memory_info()[0] / float(2 ** 20)
    return mem


def print_memory_usage(string):
    """Print the memory usage."""
    mem = get_memory_usage()
    if mpi.nb_proc > 1:
        mem = mpi.comm.allreduce(mem, op=mpi.MPI.SUM)
    if mpi.rank == 0:
        print((string + ":").ljust(30), mem, "Mo")


def print_size_in_Mo(arr, string=None):
    """Print the size of an array."""
    if string is None:
        string = "Size of ndarray (equiv. seq.)"
    else:
        string = "Size of " + string + " (equiv. seq.)"

    mem = float(arr.nbytes) * 1e-6
    if mpi.nb_proc > 1:
        mem = mpi.comm.allreduce(mem, op=mpi.MPI.SUM)
    if mpi.rank == 0:
        print(string.ljust(30), ":", mem, "Mo")


@contextlib.contextmanager
def print_options(*args, **kwargs):
    """Set print option

    example:
    >>> an_array = np.random.random((3,3))
    >>> with print_options(precision=3, suppress=True):
    >>>     print(an_array)
    """
    original = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)
    yield

    np.set_printoptions(**original)


@contextlib.contextmanager
def setenv(**env):
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards.

    Adapted from: https://gist.github.com/igniteflow/7267431

    Examples
    --------

    >>> import os
    >>> with setenv(CC="gcc"):
    >>>     print(os.getenv("CC"))  # prints: gcc
    >>> print(os.getenv("CC"))  # prints: the original value or none.
    """
    original_env = {key: os.getenv(key) for key in env}
    try:
        os.environ.update(env)
        yield
    finally:
        for key, value in original_env.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value


def config_logging(level="info", name="fluiddyn", file=None, color=False):
    """Configure a logging with a particular level and output file."""

    if not level:
        return

    if color:
        try:
            import colorlog as logging
        except ImportError:
            import logging

            print("Colored logging requires colorlog package.")
            color = False
    else:
        import logging

    level = level.lower()
    if level == "info":
        level = logging.INFO
    elif level == "debug":
        level = logging.DEBUG

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # create console handler with a higher log level
    if file is None:
        file = sys.stdout

    ch = logging.StreamHandler(file)
    ch.setLevel(level)

    # create formatter and add it to the handlers
    if color:
        formatter = logging.ColoredFormatter(
            "%(log_color)s%(levelname)s: %(message)s"
        )
    else:
        formatter = logging.Formatter("%(levelname)s: %(message)s")
    ch.setFormatter(formatter)

    # add the handlers to the logger
    logger.addHandler(ch)

    # hack to solve a "bug" in ipython notebook:
    # http://stackoverflow.com/questions/31403679/python-logging-module-duplicated-console-output-ipython-notebook-qtconsole
    logger.propagate = False



"""
Test local cluster
==================

"""
import unittest
import os
from shutil import rmtree
from .. import local
from .test_slurm_snic import SlurmTestCase, stdout_redirected


class LocalTestCase(SlurmTestCase):
    def setUp(self):
        super().setUp(local.ClusterLocal)

    def test_submit_non_default(self):
        """Test submit_script method with its non-default options."""
        self.cluster.max_walltime = "3-00:00:00"
        nb_cores_per_node = self.cluster.nb_cores_per_node // 2

        with stdout_redirected():
            self.cluster.submit_script(
                self._script,
                name_run="test",
                nb_nodes=2,
                nb_cores_per_node=nb_cores_per_node,
                walltime="2-23:59:58",
                nb_mpi_processes=1,
                retain_script=False,
                omp_num_threads=2,
                ask=False,
                bash=False,
                interactive=True,
            )


if __name__ == "__main__":
    unittest.main()



"""
IO for NS3D files (:mod:`fluiddyn.io.ns3d`)
===================================================

.. currentmodule:: fluiddyn.io.ns3d

Provides the classes :class:`NS3DFieldFile`.

.. autoclass:: NS3DFieldFile
   :members:

"""

from __future__ import division, print_function

from builtins import range
from builtins import object
import sys

import numpy as np

from .binary import BinFile


def print_with_emptyend(s):
    print(s, end="")
    sys.stdout.flush()


class NS3DFile:
    """Fields in a NS3D binary file."""

    def __init__(self, path_file=None):

        if path_file is not None:
            self.path_file = path_file
            self.byteorder = self._get_byte_order()
            self._read_header()

    def _get_byte_order(self):
        with BinFile(self.path_file, byteorder="little") as f:
            trash, flag = f.readt(2, "I")
        if flag == 1:
            return "little"

        with BinFile(self.path_file, byteorder="big") as f:
            trash, flag = f.readt(2, "I")
        if flag == 1:
            return "big"

        else:
            raise ValueError("This file does not look like a ns3d file.")


class NS3DFieldFile(NS3DFile):
    """Fields in a NS3D binary file."""

    nb_bytes_header = 148

    def _read_header(self):
        """Read the header of the file."""

        with BinFile(self.path_file, byteorder=self.byteorder) as f:
            trash, flag, self.nx, self.ny, self.nz = f.readt(5, "uint32")
            self.lx, self.ly, self.lz, self.dt = f.readt(4, "float64")

            self.truncate, self.type_trunc = f.readt(2, "uint32")
            (self.rtrunc_x, self.rtrunc_y, self.rtrunc_z, self.nu) = f.readt(
                4, "float64"
            )

            self.stratification = f.readt(1, "uint32")
            self.N, self.schm, self.omega2 = f.readt(3, "float64")

            self.perturb, self.lin, trash, trash = f.readt(4, "uint32")

            self.time = f.readt(1, "float64")

        self.shape = (self.nx, self.ny, self.nz)
        if not self.stratification:
            self.N == 0

        self.Re = np.round(1.0 / self.nu, decimals=2)
        self.Fh = np.round(1.0 / self.N, decimals=4)

    def read_field(self, ifield=0):
        nb_pts_one_field = self.nx * self.ny * self.nz
        field = np.empty([self.nz, self.ny, self.nx])
        with BinFile(self.path_file, byteorder=self.byteorder) as f:
            f.seek(self.nb_bytes_header + ifield * (nb_pts_one_field + 1) * 8 + 4)
            for iz in range(self.nz):
                field[iz] = np.array(
                    f.readt(self.nx * self.ny, "float64")
                ).reshape([self.ny, self.nx])
        return field

    def read_xy(self, ifield=0, iz=0):

        nb_pts_one_field = self.nx * self.ny * self.nz

        with BinFile(self.path_file, byteorder=self.byteorder) as f:
            f.seek(
                self.nb_bytes_header
                + ifield * (nb_pts_one_field + 1) * 8
                + 4
                + self.nx * self.ny * iz * 8
            )
            field = np.array(f.readt(self.nx * self.ny, "float64"))

        return field.reshape([self.ny, self.nx])

    def save_with_byteorder_changed(self):

        if self.byteorder.startswith("little"):
            newbyteorder = "big"
        elif self.byteorder.startswith("big"):
            newbyteorder = "little"
        else:
            raise ValueError("byteorder should start with little or big.")

        new_path = self.path_file + "_" + newbyteorder + "-endian"

        nb_pts = self.nx * self.ny * self.nz

        with BinFile(new_path, "w", byteorder=newbyteorder) as f:
            # write the header
            f.write_as([124, 1, self.nx, self.ny, self.nz], "uint32")
            f.write_as([self.lx, self.ly, self.lz, self.dt], "float64")
            f.write_as([self.truncate, self.type_trunc], "uint32")
            f.write_as(
                [self.rtrunc_x, self.rtrunc_y, self.rtrunc_z, self.nu], "float64"
            )
            f.write_as(self.stratification, "uint32")
            f.write_as([self.N, self.schm, self.omega2], "float64")
            f.write_as([self.perturb, self.lin, 124], "uint32")
            # write the time
            f.write_as(8, "uint32")
            f.write_as(self.time, "float64")
            f.write_as(8, "uint32")
            # write the 6 fields
            for ifield in range(7):
                try:
                    field = self.read_field(ifield)
                except ValueError:
                    break

                f.write_as(nb_pts, "uint32")
                f.write_as(field.flatten(), "float64")
                f.write_as(nb_pts, "uint32")

        print("New file saved:\n" + new_path)

    def save_with_resol_changed(self, nx_new, ny_new, nz_new):
        from ..calcul import easypyfft

        print_with_emptyend("init. FFTW3DReal2Complex for input file...")
        op = easypyfft.FFTW3DReal2Complex(self.nx, self.ny, self.nz)
        print(" Done.")
        print_with_emptyend("init. FFTW3DReal2Complex for output file...")
        op_new = easypyfft.FFTW3DReal2Complex(nx_new, ny_new, nz_new)
        print(" Done.")

        new_path = self.path_file + f"_{nx_new}x{ny_new}x{nz_new}"

        nb_pts = nx_new * ny_new * nz_new

        with BinFile(new_path, "w", byteorder=self.byteorder) as f:
            # write the header
            f.write_as([124, 1, nx_new, ny_new, nz_new], "uint32")
            f.write_as([self.lx, self.ly, self.lz, self.dt], "float64")
            f.write_as([self.truncate, self.type_trunc], "uint32")
            f.write_as(
                [self.rtrunc_x, self.rtrunc_y, self.rtrunc_z, self.nu], "float64"
            )
            f.write_as(self.stratification, "uint32")
            f.write_as([self.N, self.schm, self.omega2], "float64")
            f.write_as([self.perturb, self.lin, 124], "uint32")
            # write the time
            f.write_as(8, "uint32")
            f.write_as(self.time, "float64")
            f.write_as(8, "uint32")
            # write the 4 fields
            for ifield in range(4):
                print(f"treat field {ifield} (over 4)")
                print_with_emptyend("    read_field...")
                try:
                    field = self.read_field(ifield)
                except ValueError:
                    break

                print_with_emptyend(" Done.\n    _compute_field_new_resol...")
                field_new = self._compute_field_new_resol(field, op, op_new)
                del (field)
                print_with_emptyend(
                    " Done.\n    write the array in the new file..."
                )
                f.write_as(nb_pts, "uint32")
                f.write_as(field_new.flat, "float64")
                del (field_new)
                print(" Done.")
                f.write_as(nb_pts, "uint32")

        print("New file saved:\n" + new_path)

    def _compute_field_new_resol(self, field, op, op_new):

        nz_new, ny_new, nx_new = op_new.shapeK

        nx_min = min(self.nx, nx_new)
        ny_min = min(self.ny, ny_new)
        nz_min = min(self.nz, nz_new)

        f_Fourier = op.fft(field)
        f_Fourier_new = np.zeros(op_new.shapeK, dtype=np.complex128)

        nkx = nx_min // 2 + 1
        # for z and y, we take advantage of the "from the end" Python
        # index notation:
        for iz in [0, nz_min // 2]:
            for iy in [0, ny_min // 2]:
                for ix in range(nkx):
                    f_Fourier_new[iz, iy, ix] = f_Fourier[iz, iy, ix]
            for iy in range(1, ny_min // 2):
                for ix in range(nkx):
                    f_Fourier_new[iz, iy, ix] = f_Fourier[iz, iy, ix]
                    f_Fourier_new[iz, -iy, ix] = f_Fourier[iz, -iy, ix]
        for iz in range(1, nz_min // 2):
            for iy in [0, ny_min // 2]:
                for ix in range(nkx):
                    f_Fourier_new[iz, iy, ix] = f_Fourier[iz, iy, ix]
                    f_Fourier_new[-iz, iy, ix] = f_Fourier[-iz, iy, ix]
            for iy in range(1, ny_min // 2):
                for ix in range(nkx):
                    f_Fourier_new[iz, iy, ix] = f_Fourier[iz, iy, ix]
                    f_Fourier_new[iz, -iy, ix] = f_Fourier[iz, -iy, ix]
                    f_Fourier_new[-iz, iy, ix] = f_Fourier[-iz, iy, ix]
                    f_Fourier_new[-iz, -iy, ix] = f_Fourier[-iz, -iy, ix]

        return op_new.ifft(f_Fourier_new)


class NS3DForcingInfoFile(NS3DFile):
    """Information on forcing NS3D binary file."""

    def _read_header(self):
        """Read the header of the file."""

        with BinFile(self.path_file, byteorder=self.byteorder) as f:
            trash, flag = f.readt(2, "uint32")
            self.lx, self.ly, self.Delta_t = f.readt(3, "float64")
            (
                self.nb_fields,
                self.nb_Delta_t,
                self.nkx,
                self.nky,
                trash,
                trash2,
            ) = f.readt(6, "uint32")
            self.vec_ind_field = f.readt(self.nb_Delta_t, "uint32")

    def save_with_byteorder_changed(self):

        if self.byteorder.startswith("little"):
            newbyteorder = "big"
        elif self.byteorder.startswith("big"):
            newbyteorder = "little"
        else:
            raise ValueError("byteorder should start with little or big.")

        new_path = self.path_file + "_" + newbyteorder + "-endian"

        with BinFile(new_path, "w", byteorder=newbyteorder) as f:
            f.write_as([44, 1], "uint32")
            f.write_as([self.lx, self.ly, self.Delta_t], "float64")
            f.write_as(
                [
                    self.nb_fields,
                    self.nb_Delta_t,
                    self.nkx,
                    self.nky,
                    44,
                    self.nb_Delta_t * 4,
                ],
                "uint32",
            )
            f.write_as(self.vec_ind_field, "uint32")
            f.write_as(self.nb_Delta_t * 4, "uint32")

        print("New file saved:\n" + new_path)


class NS3DForcingSpectralFile:
    """Fields in a NS3D binary file."""

    def __init__(self, path_file):
        if "forcing_2D_spectral.in" not in path_file:
            raise ValueError('"forcing_2D_spectral.in" should be in path_file.')

        base, suffix = tuple(path_file.split("forcing_2D_spectral.in"))
        path_file_info = base + "forcing_2D_info.in" + suffix
        f_info = NS3DForcingInfoFile(path_file_info)

        self.path_file = path_file
        self.byteorder = f_info.byteorder
        self.nkx = f_info.nkx
        self.nky = f_info.nky
        self.nb_fields = f_info.nb_fields

    def read_one_forcing_field(self, iforcing):

        if iforcing < 0 or iforcing > self.nb_fields - 1:
            raise ValueError("iforcing should be >=0 and <self.nb_fields-1.")

        with BinFile(self.path_file, byteorder=self.byteorder) as f:
            f.seek(3 * iforcing * 2 * self.nkx * self.nky * 8)
            tdata = f.readt(3 * 2 * self.nkx * self.nky, "float64")

        return tdata

    def save_with_byteorder_changed(self):

        if self.byteorder.startswith("little"):
            newbyteorder = "big"
        elif self.byteorder.startswith("big"):
            newbyteorder = "little"
        else:
            raise ValueError("byteorder should start with little or big.")

        new_path = self.path_file + "_" + newbyteorder + "-endian"

        with BinFile(new_path, "w", byteorder=newbyteorder) as f:

            for iforcing in range(self.nb_fields):
                tdata = self.read_one_forcing_field(iforcing)
                f.write_as(tdata, "float64")

        print("New file saved:\n" + new_path)


if __name__ == "__main__":

    path_dir = (
        "/home/users/augier3pi/useful/project/14LADHYX/NS3D_results_froggy/"
        "ns3d_exp_HYPER_288x288x96_L=30x30x10_"
        "nu=0.0001266_N=2.0_Tend=400_2015-01-09_19-18-52"
    )

# ff0 = NS3DFieldFile(path_file=path_dir +
#                     '/velo_rho_vort.t=0400.045')

# ff1 = NS3DFieldFile(path_file=path_dir +
#                     '/velo_rho_vort.t=0400.045_64x64x32')

# ff2 = NS3DFieldFile(path_file=path_dir +
#                     '/velo_rho_vort.t=0005.010_128x128x32')



"""
Test rdvision module
====================

"""

import unittest
import os
from glob import glob
from shutil import rmtree, copy

from ..rdvision import read_seq, read_xml, read_sqb, SetOfFiles

input_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "rdvision_files")
)
input_files = glob(os.path.join(input_dir, "*"))


class TestDantex(unittest.TestCase):
    """Test fluiddyn.io.rdvision module."""

    @classmethod
    def setUpClass(cls):
        work_dir = cls._work_dir = "test_fluiddyn_io_rdvision"
        if not os.path.exists(work_dir):
            os.mkdir(work_dir)

        for path in input_files:
            copy(path, work_dir)

        cls.base_name = os.path.splitext(input_files[0])[0]

        os.chdir(work_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir("..")
        rmtree(cls._work_dir)

    def test(self):
        read_xml(self.base_name)
        read_sqb(self.base_name)
        SetOfFiles(self.base_name)


if __name__ == "__main__":
    unittest.main()



import unittest
import os
import sys

from fluiddyn.io import stdout_redirected
from fluiddyn.util import mpi

from .. import main

here = os.path.abspath(os.path.dirname(__file__))


class TestsBench(unittest.TestCase):
    def test2d(self):
        with stdout_redirected():
            if mpi.rank > 0:
                return

            path = os.path.join(here, "courant.m")
            args = ["fluidmat2py ", path]
            sys.argv = args
            main()


if __name__ == "__main__":
    unittest.main()



"""Ciment clusters (:mod:`fluiddyn.clusters.ciment`)
====================================================

Provides:

.. autoclass:: Froggy
   :members:

`CIMENT <https://ciment.ujf-grenoble.fr>`_ is the Grenoble University
High Performance Computing (HPC) centre.

"""
from sys import version_info as version

from fluiddyn.clusters.oar import ClusterOAR


class Froggy(ClusterOAR):
    name_cluster = "froggy"
    has_to_add_name_cluster = False
    nb_cores_per_node = 16
    frontends = [""]

    def __init__(self):

        super().__init__()

        self.commands_setting_env = [
            "source /applis/site/env.bash",
            "export MODULEPATH="
            "/home/PROJECTS/pr-stratturb/modulefiles:$MODULEPATH",
            "module load python/{}.{}.{}".format(
                version.major, version.minor, version.micro
            ),
            "source /home/$USER/opt/mypy{}.{}/bin/activate".format(
                version.major, version.minor
            ),
        ]



"""IO for python files containing data (:mod:`fluiddyn.io.in_py`)
=================================================================

.. autofunction:: save_in_py

"""

import sys
import numpy as np


def save_in_py(path, variables, names=None):
    """Save data in `variables` in the file `path`.

    Parameters
    ----------

    path : str

      Path of the file where the data are saved (has to end with '.py').

    variables : dict

      Contains the variables to be saved.

    names : None or sequence of str.

      If None, all variables in variables are saved, else, only the variables
      with name in names of the variables to be saved.

    Examples
    --------

    .. code::

       a = 1
       b = 'str'
       c = np.ones(2)

       save_in_py('myfile.py', locals(), ('a', 'b', 'c'))

    or from a dictionary::

       d = {'a': 1, 'b': 'str', 'c': np.ones(2)}
       save_in_py('myfile.py', d)

    """
    if not path.endswith(".py"):
        raise ValueError

    if names is not None:
        variables_all = variables
        variables = {}
        for name in names:
            variables[name] = variables_all[name]

    np.set_printoptions(threshold=sys.maxsize)
    with open(path, "w") as f:
        f.write("from numpy import array, nan\n")
        for key, value in variables.items():
            f.write(key + " = " + repr(value) + "\n")



"""
Clocking timer (:mod:`fluiddyn.util.timer`)
===========================================

.. currentmodule:: fluiddyn.util.timer

Provides:

.. autoclass:: Timer
   :members:
   :private-members:

.. autoclass:: TimerIrregular
   :members:
   :private-members:

"""

from __future__ import division, print_function

from builtins import range
from builtins import object
import time
from datetime import timedelta
import operator


def parse_timestamp(timestr):
    """Converts a timestamp to a time.struct_time object."""
    time_formats = ["%d-%H:%M:%S", "%H:%M:%S", "%M:%S", "%S"]

    for tf in time_formats:
        try:
            time_struct = time.strptime(timestr, tf)
        except ValueError:
            continue

        else:
            break

    else:
        raise ValueError(
            "Timestamp should be in one of the following "
            "formats: {}".format(time_formats)
        )

    return time_struct


def timestamp_to_seconds(timestr):
    """Converts a timestamp to total seconds."""
    ts = parse_timestamp(timestr)
    # Handle both formats:
    days = ts.tm_yday if "-" in timestr else (ts.tm_yday - 1)
    td = timedelta(
        hours=ts.tm_hour, minutes=ts.tm_min, seconds=ts.tm_sec, days=days
    )

    return int(td.total_seconds())


def time_gteq(timestr1, timestr2):
    """Compares two timestamps strings."""
    return TimeStr(timestr1) >= TimeStr(timestr2)


class TimeStr(str):
    """String types with special comparison operators to compare equivalent
    time.struct.

    """

    def __init__(self, value):
        self._struct = parse_timestamp(value)
        super().__init__()

    def _operate(self, operator_func, other):
        if isinstance(other, TimeStr):
            return operator_func(self._struct, other._struct)

        elif isinstance(other, str):
            return operator_func(self._struct, parse_timestamp(other))

    def __le__(self, other):
        return self._operate(operator.le, other)

    def __lt__(self, other):
        return self._operate(operator.lt, other)

    def __eq__(self, other):
        return self._operate(operator.eq, other)

    def __ne__(self, other):
        return self._operate(operator.ne, other)

    def __ge__(self, other):
        return self._operate(operator.ge, other)

    def __gt__(self, other):
        return self._operate(operator.gt, other)


class Timer:
    """Timer ticking with a particular period.

    Parameters
    ----------

    time_between_ticks : float
       Period between the ticks (in s).

    """

    def __init__(self, time_between_ticks):
        self.time_between_ticks = time_between_ticks
        self.tstart = time.time()
        self.last_period = -1

    def wait_tick(self):
        """Block till the next tick."""
        tnow = time.time()
        tsleep = (
            self.time_between_ticks
            - (tnow - self.tstart) % self.time_between_ticks
        )
        this_period = int((tnow - self.tstart) / self.time_between_ticks)
        if this_period == self.last_period:
            self.last_period = this_period + 1
            tsleep += self.time_between_ticks
        else:
            self.last_period = this_period

        time.sleep(tsleep)
        return time.time() - self.tstart

    def restart(self):
        self.tstart = time.time()

    def get_time_till_start(self):
        return time.time() - self.tstart


class TimerIrregular(Timer):
    """Timer ticking for a numpy array of time.

    Parameters
    ----------

    times_ticks: (sorted) sequence
      A sequence of times (in second), which can be irregular.

    """

    def __init__(self, times_ticks):
        t0 = times_ticks[0]
        self.times_ticks = [ti - t0 for ti in times_ticks[1:]]
        self.tstart = time.time()

    def wait_tick(self):
        """Block till the next tick."""
        tnow = time.time() - self.tstart
        if self.times_ticks:
            while self.times_ticks and self.times_ticks[0] - tnow < 0:
                self.times_ticks = self.times_ticks[1:]
        if self.times_ticks:
            tsleep = self.times_ticks[0] - tnow
            self.times_ticks = self.times_ticks[1:]
            time.sleep(tsleep)
        return time.time() - self.tstart


if __name__ == "__main__":

    tstart = time.time()

    timer = Timer(2)
    for it in range(10):
        # print(it, end=' ')

        # time.sleep(float(it)/5)

        # print('before timer.wait_tick()', time.time()-timer.tstart)
        timer.wait_tick()
    # print('after  timer.wait_tick()', time.time()-timer.tstart)

    tend = time.time()

    print(tend - tstart)



"""
Spherical harmonics operators (:mod:`fluiddyn.calcul.sphericalharmo`)
=====================================================================

This module requires the C library shtns. To install it, run something like::

  ./configure --enable-openmp --enable-python
  make && make install

.. autoclass:: EasySHT
   :members:
   :private-members:

"""


from __future__ import division, print_function


import sys
from time import time
from warnings import warn


import numpy as np

try:
    import shtns
except ImportError:
    warn(
        "ImportError shtns.\n\n"
        "To install shtns, you can run the following:\n"
        "    hg clone https://bitbucket.org/nschaeff/shtns\n"
        "    cd shtns\n"
        "    ./configure --prefix=$VIRTUAL_ENV --enable-python\n"
        "    make\n"
        "    make install\n"
    )
else:
    sht_orthonormal = shtns.sht_orthonormal
    sht_fourpi = shtns.sht_fourpi
    sht_schmidt = shtns.sht_schmidt

    # print('sht_norms', sht_orthonormal, sht_fourpi, sht_schmidt)

    sht_gauss = shtns.sht_gauss
    sht_auto = shtns.sht_auto
    sht_reg_fast = shtns.sht_reg_fast
    sht_reg_dct = shtns.sht_reg_dct
    sht_quick_init = shtns.sht_quick_init
    sht_reg_poles = shtns.sht_reg_poles
    sht_gauss_fly = shtns.sht_gauss_fly

    SHT_THETA_CONTIGUOUS = shtns.SHT_THETA_CONTIGUOUS

    SHT_PHI_CONTIGUOUS = shtns.SHT_PHI_CONTIGUOUS

    SHT_NO_CS_PHASE = shtns.SHT_NO_CS_PHASE
    SHT_REAL_NORM = shtns.SHT_REAL_NORM
    SHT_SCALAR_ONLY = shtns.SHT_SCALAR_ONLY
    SHT_SOUTH_POLE_FIRST = shtns.SHT_SOUTH_POLE_FIRST


"""
flags for normalization (norm = ...)
 sht_orthonormal = 0  # orthonormalized spherical harmonics (default).
 sht_fourpi      = 1  # Geodesy and spectral analysis : 4.pi normalization.
 sht_schmidt     = 2  # Schmidt semi-normalized : 4.pi/(2l+1)

flags for spatial data layouts and grids used by SHTns
use the bitwise operator OR "|" to separate the different flags
for example:
flags = (shtns.sht_quick_init | shtns.SHT_PHI_CONTIGUOUS |
         shtns.SHT_SOUTH_POLE_FIRST)
"""

radius_earth = 6367470.0  # earth radius (m)


def compute_lmax(nlat, nl_order=2):
    return int(2 * nlat / (nl_order + 1) - 1)


def compute_nlatnlon(lmax, nl_order=2):
    if lmax % 2 == 0:
        lmax += 1
    nlat = int((lmax + 1) * (nl_order + 1) / 2)
    if nlat % 2 == 1:
        nlat += 1
    return nlat, 2 * nlat


def _for_test(i):
    if i % 2 == 1:
        assert i == compute_lmax(compute_nlatnlon(i)[0])


def bin_int(n, width):
    return "".join(str((n >> i) & 1) for i in range(width - 1, -1, -1))


class EasySHT:
    """Simple possibilities of shtns.

    Less possibilities but very simple to use...  It has been written
    specially for atmospheric applications.

    Creation of a default instance::

      esh = EasySHT(lmax=15)

    Parameters
    ----------

    lmax : number {15}
        Truncation degree.

    mmax : {None, int}
        If is None, triangular truncation.

    mres : 1
        Azimutal symmetry coefficient (see shtns documentation).

    norm=shtns.sht_fourpi
        For SH with quadratic mean equal to unity.

    nlat : {None, int}
        If None, computed by shtns to avoid aliasing.

    nlon : {None, int}
        If None, computed by shtns to avoid aliasing.

    flags : {sht_quick_init|SHT_PHI_CONTIGUOUS|SHT_SOUTH_POLE_FIRST, int}
        Option flag for shtns.

    polar_opt : {1.0e-8, float}
        Polar optimization threshold.

    nl_order : {2, int}
        Nonlinear order of the equations (used to compute nlat and nlon).

    radius : {radius_earth, number}
        Radius of the sphere (in meters)

    Notes
    -----

    In contrast as with shtns, with easypysht the meridional unit
    vector points towards the North if shtns.SHT_SOUTH_POLE_FIRST is
    used (this is the default) and it points towards the South if
    shtns.SHT_SOUTH_POLE_FIRST is not given in flags (thus there is a
    change of sign in the meridional velocity).

    easypysht has been written for atmospheric applications thus some
    usual notations are used.

    Here are some definitions useful to understand the code:

    - l     denotes the degree of the spherical harmonic functions
    - m     denotes the order of the spherical harmonic functions
    - SH    denotes spherical harmonic (spectral array)
    - spat  denotes spatial array

    - lat   denotes latitude
    - lon   denotes longitude
    - kh    denotes horizontal wavenumber
    - u     denotes longitudinal velocity
    - v     denotes meridional velocity (the sign depends on the used base)

    - hdiv  denotes horizontal divergence
    - hrot  denotes vertical vorticity (curl on the sphere)
    - grad  denotes the horizontal gradient

    Variables and functions about spectral space array:
    lmax, mmax, mres, nlm, idx_lm(), l_idx, m_idx, l2_idx

    Variables about grid and real space array:
    nlat, nlon, lats, lons, sin_lats, LATS, LONS

    Variables for spectra:
    l2_l, kh_l

    self.sh is the instance of the class sht defined in shtns.py.
    All functions and variables of this class can be used directly
    from this instance, for example::

      esh.sh.sh_to_point(f_lm, cost, phi)

    where t is the colatitude and phi is the longitude.

    """

    def __init__(
        self,
        lmax=15,
        mmax=None,
        mres=1,
        norm=None,
        nlat=None,
        nlon=None,
        flags=None,
        polar_opt=1.0e-8,
        nl_order=2,
        radius=radius_earth,
    ):

        # to get a clear ImportError
        import shtns

        if norm is None:
            norm = sht_fourpi

        if flags is None:
            flags = (
                shtns.sht_quick_init
                | shtns.SHT_PHI_CONTIGUOUS
                | shtns.SHT_SOUTH_POLE_FIRST
            )

        # print(lmax,mmax,mres,nlat,nlon,flags,polar_opt,nl_order,radius)
        # print('flags', flags,
        #       shtns.sht_quick_init, shtns.SHT_PHI_CONTIGUOUS,
        #       shtns.SHT_SOUTH_POLE_FIRST)

        if lmax is None and nlat is None:
            raise ValueError("lmax or nlat should be given.")

        elif lmax is None:
            lmax = compute_lmax(nlat)

        # print('lmax', lmax)

        self.lmax = int(lmax)
        self.radius = float(radius)
        # print(self.radius, radius_earth)

        if mmax is None and mres == 1:
            # triangular truncation
            self.mmax = self.lmax
            self.mres = 1
        else:
            self.mmax = int(float(lmax) / mres)
            self.mres = mres

        print(
            "lmax={}, mmax={}, mres={}, norm={}".format(
                self.lmax, self.mmax, self.mres, norm
            )
        )

        self.sh = shtns.sht(self.lmax, self.mmax, self.mres, norm=norm)

        bin_flags = bin_int(flags, 14)
        # print(bin_flags, bin_flags[-14])
        if bin_flags[-14] == "1":
            self.order_lat = "south_to_north"
        else:
            self.order_lat = "north_to_south"

        # in shtns, 0 means None
        if nlat is None:
            nlat = 0
        if nlon is None:
            nlon = 0

        self.nlat, self.nlon = self.sh.set_grid(
            nlat=nlat,
            nphi=nlon,
            flags=flags,
            polar_opt=polar_opt,
            nl_order=nl_order,
        )
        # print('flags', flags)

        self.sin_lats = self.sh.cos_theta
        self.lons = np.arange(self.nlon) * 360.0 / (self.nlon * self.mres)
        self.lats = np.arcsin(self.sin_lats) / np.pi * 180.0

        # for fluidsim plotting
        self.x_seq = self.lons
        self.y_seq = self.lats

        # degree
        self.l_idx = self.sh.l
        self.l2_idx = self.l_idx * (self.l_idx + 1)
        # laplacian:=l(l+1)/r^2 and laplacian^2
        self.K2 = self.l2_idx / self.radius ** 2
        self.K4 = self.K2 ** 2
        self.K8 = self.K4 ** 2

        self.K2_not0 = self.K2[:]
        COND = self.l2_idx == 0
        self.K2_not0[COND] = 1e-15

        # print('l2_idx=', self.l2_idx)
        self.m_idx = self.sh.m
        self.nlm = self.sh.nlm

        self.deltax = 360.0 / (self.nlon * self.mres)
        # wrong but useful for fluidsim
        self.deltay = self.deltax

        # create arrays 2D lats et lons
        self.LONS, self.LATS = np.meshgrid(self.lons, self.lats)
        self.cosLATS = np.cos(self.LATS / 180 * np.pi)

        self.lrange = np.arange(self.lmax + 1)
        self.l2_l = self.lrange * (self.lrange + 1)
        self.kh_l = np.sqrt(self.l2_l) / self.radius

        self._complex64_save_netCFD = np.dtype(
            [("real", np.float32), ("imag", np.float32)]
        )

        # print(lmax,mmax,mres,nlat,nlon,flags,polar_opt,nl_order,radius)

        # no MPI decomposition
        self.shapeX = self.shapeX_loc = self.shapeX_seq = (self.nlat, self.nlon)
        self.shapeK = self.shapeK_loc = self.shapeK_seq = (self.nlm,)

        self.sht_as_arg = self.sh.spat_to_SH
        self.isht_as_arg = self.sh.SH_to_spat

        self._zeros_sh = self.create_array_sh(0.0)

    def vec_from_rotsh(self, rot_sh):
        return self.uv_from_hdivrotsh(self._zeros_sh, rot_sh)

    def vec_from_divsh(self, div_sh):
        return self.uv_from_hdivrotsh(div_sh, self._zeros_sh)

    def produce_str_describing_oper(self):
        """Produce a string describing the operator."""
        return f"lmax{self.lmax}_nlat{self.nlat}_nlon{self.nlon}"

    def produce_long_str_describing_oper(self):
        return "\n".join(
            (
                "Spherical harmonic transforms "
                "nlat = {} ; nlon = {}".format(self.nlat, self.nlon),
                "(1 point every {:6.2g} m for the current sphere of radius {})".format(
                    2 * np.pi * self.radius / self.nlon, self.radius
                ),
                "(1 point every {:6.2g} km for the earth)\n".format(
                    2 * np.pi * radius_earth / self.nlon / 1000
                ),
            )
        )

    def idx_lm(self, l, m):
        """ idx_lm(self, l,m)"""
        if l >= 0 and 0 <= m <= l:
            return self.sh.idx(int(l), int(m))

        else:
            raise ValueError("not (l>=0 and m>=0 and m<=l)")

    # functions for initialisation of field

    def create_array_sh(self, value=None, dtype=complex):
        """Create an array representing a field in spectral space."""
        if value is None:
            field_lm = np.empty(self.nlm, dtype)
        elif value == "rand":
            field_lm = np.random.randn(self.nlm) + 1.0j * np.random.randn(
                self.nlm
            )
        elif value == 0:
            field_lm = np.zeros(self.nlm, dtype)
        else:
            field_lm = value * np.ones(self.nlm, dtype)
        return field_lm

    def create_array_spat(self, value=None):
        """Create an array representing a field in spatial space."""
        if value is None:
            field = np.empty(self.shapeX)
        elif value == "rand":
            field = np.random.randn(self.nlat, self.nlon)
        elif value == 0:
            field = np.zeros(self.shapeX)
        else:
            field = value * np.ones(self.shapeX)
        # a spatial array matching a grid build with SHT_PHI_CONTIGUOUS
        return field

    def convert2complex64_save_netCFD(self, f_lm):
        result = np.empty(f_lm.shape, self._complex64_save_netCFD)
        result["real"] = f_lm.real
        result["imag"] = f_lm.imag
        return result

    # functions for scalar spherical harmonic transforms (forward and backward)

    def sht(self, field):
        field_lm = self.create_array_sh()
        self.sh.spat_to_SH(field, field_lm)
        return field_lm

    def isht(self, field_lm):
        field = self.create_array_spat()
        self.sh.SH_to_spat(field_lm, field)
        return field

    def sh_from_spat(self, field, field_lm=None):
        """Spherical harmonic transform.

        examples:
        f_lm = sh_from_spat(f)

        or if f_lm already exists:
        sh_from_spat(f, f_lm)
        """
        if field.dtype != np.dtype("float"):
            field = np.array(field, float)
        if field_lm is None:
            field_lm = self.create_array_sh()

        self.sh.spat_to_SH(field, field_lm)
        return field_lm

    def spat_from_sh(self, field_lm, field=None):
        """Inverse spherical harmonic transform.

        """
        if field_lm.dtype != np.dtype("complex"):
            field_lm = np.array(field_lm, complex)
        if field is None:
            field = self.create_array_spat()
        self.sh.SH_to_spat(field_lm, field)
        return field

    def chrono_sht(self, nb_sht=10):
        """Microbenchmark forward and inverse SHT, and 
        vorticity, divergence <-> u, v transformations.

        """
        f = np.random.rand(self.nlat, self.nlon)
        t1 = time()
        for i in range(nb_sht):
            f_lm = self.sh_from_spat(f)
        t2 = time()
        print(
            "    mean time for 1 forward SHT: {:3.6f} s".format(
                (t2 - t1) / nb_sht
            )
        )
        t1 = time()
        for i in range(nb_sht):
            f = self.spat_from_sh(f_lm)
        t2 = time()
        print(
            "    mean time for 1 inverse SHT: {:3.6f} s".format(
                (t2 - t1) / nb_sht
            )
        )
        uu = np.random.rand(self.nlat, self.nlon)
        vv = np.random.rand(self.nlat, self.nlon)
        t1 = time()
        for i in range(nb_sht):
            hdiv, hrot = self.hdivrotsh_from_uv(uu, vv)
        t2 = time()
        print(
            "    mean time for hdivrotsh_from_uuvv vectorial SHT: "
            "{:3.6f} s".format((t2 - t1) / nb_sht)
        )
        t1 = time()
        for i in range(nb_sht):
            uu, vv = self.uv_from_hdivrotsh(hdiv, hrot)
        t2 = time()
        print(
            "    mean time for uv_from_hdivrotsh vectorial SHT: "
            "{:3.6f} s".format((t2 - t1) / nb_sht)
        )

    # functions for 2D vectorial spherical harmonic transforms

    def uv_from_hdivrotsh(self, hdiv_lm, hrot_lm, uu=None, vv=None):
        """
        u, v from div, rot (u and v are overwritten)
        """
        if uu is None:
            uu = self.create_array_spat()
            vv = self.create_array_spat()
        uD_lm = self.create_array_sh(0.0)
        uR_lm = self.create_array_sh(0.0)
        COND = self.l2_idx > 0
        uD_lm[COND] = -hdiv_lm[COND] / self.l2_idx[COND] * self.radius
        uR_lm[COND] = -hrot_lm[COND] / self.l2_idx[COND] * self.radius
        self.sh.SHsphtor_to_spat(uD_lm, uR_lm, vv, uu)
        # if self.order_lat == 'south_to_north':
        #    vv[:] = -vv+0       # because SHTns uses colatitude basis
        return uu, vv

    def hdivrotsh_from_uv(self, uu, vv, hdiv_lm=None, hrot_lm=None):
        """Compute hdivrotsh from uuvv.

        (div_lm and rot_lm are overwritten)
        """
        if hdiv_lm is None:
            hdiv_lm = self.create_array_sh()
            hrot_lm = self.create_array_sh()

        # if self.order_lat == 'south_to_north':
        #     vv = -vv
        # print('order_lat',self.order_lat)
        self.sh.spat_to_SHsphtor(vv, uu, hdiv_lm, hrot_lm)
        # in fact there is uD_lm in hdiv_lm and
        #                  uR_lm in hrot_lm
        # we compute div_lm and rot_lm
        hdiv_lm[:] = -self.l2_idx * hdiv_lm[:] / self.radius
        # removed minus
        hrot_lm[:] = -self.l2_idx * hrot_lm[:] / self.radius
        return hdiv_lm, hrot_lm

    def uv_from_uDuRsh(self, uD_lm, uR_lm, uu=None, vv=None):
        """Compute velocities uu, vv from uD, uR (uu and vv are 
        overwritten).

        """
        if uu is None:
            uu = self.create_array_spat()

        if vv is None:
            vv = self.create_array_spat()
        self.sh.SHsphtor_to_spat(uD_lm, uR_lm, vv, uu)

        # if self.order_lat == 'south_to_north':
        #    vv[:] = -vv+0       # because SHTns uses colatitude basis

        return uu, vv

    def uDuRsh_from_uv(self, uu, vv, uD_lm=None, uR_lm=None):
        """Compute helmholtz decomposition of the velocities from uu, vv.
        (uD_lm and uR_lm are overwritten).

        """
        if uD_lm is None:
            uD_lm = self.create_array_sh()

        if uR_lm is None:
            uR_lm = self.create_array_sh()

        # if self.order_lat == 'south_to_north':
        #     vv = -vv
        # print('order_lat', self.order_lat)
        self.sh.spat_to_SHsphtor(vv, uu, uD_lm, uR_lm)
        # removed minus
        uD_lm[:] = -uD_lm[:]
        uR_lm[:] = -uR_lm[:]
        # print(self.radius)
        return uD_lm, uR_lm

    def hdivrotsh_from_uDuRsh(self, uD_lm, uR_lm, hdiv_lm=None, hrot_lm=None):
        """Compute horizontal divergence and vertical vorticity spherical
        harmonics from velocity vector spherical harmonics (hdiv_lm and hrot_lm
        are overwritten).

        """
        if hdiv_lm is None:
            hdiv_lm = self.create_array_sh(0.0)  # TODO: Maybe can be empty?

        if hrot_lm is None:
            hrot_lm = self.create_array_sh(0.0)

        hdiv_lm[:] = -self.l2_idx * uD_lm / self.radius
        hrot_lm[:] = self.l2_idx * uR_lm / self.radius
        return hdiv_lm, hrot_lm

    def uDuRsh_from_hdivrotsh(self, hdiv_lm, hrot_lm, uD_lm=None, uR_lm=None):
        """Compute velocity vector spherical harmonics from horizontal
        divergence and vertical vorticity spherical harmonics (uD_lm and
        uR_lm are overwritten).

        """
        if uD_lm is None:
            uD_lm = self.create_array_sh(0.0)

        if uR_lm is None:
            uR_lm = self.create_array_sh(0.0)

        COND = self.l2_idx > 0
        uD_lm[COND] = -hdiv_lm[COND] / self.l2_idx[COND] * self.radius
        uR_lm[COND] = +hrot_lm[COND] / self.l2_idx[COND] * self.radius
        # print(self.radius)
        return uD_lm, uR_lm

    def gradf_from_fsh(self, f_lm, gradf_lon=None, gradf_lat=None):
        """gradf from fsh.

        Compute the gradient of a function f from its spherical
        harmonic coeff f_lm (gradf_lon and gradf_lat are overwritten)

        """
        if gradf_lon is None:
            gradf_lon = self.create_array_spat(0)
            gradf_lat = self.create_array_spat(0)  # becareful bug if not 0!!!
        #       We do not use SHsph_to_spat() because it seems that there is a
        #       problem
        # FIXME: av: What exactly is the problem and does it still exist? Issue link?
        ####    self.sh.SHsph_to_spat(f_lm, gradf_lat, gradf_lon)
        #       instead we use SHsphtor_to_spat(...) with tor_lm= zeros_lm
        # zeros_lm = self.create_array_sh(0.)
        self.sh.SHsphtor_to_spat(f_lm, self._zeros_sh, gradf_lat, gradf_lon)

        # if self.order_lat == 'south_to_north':
        #    sign_inv_vv = -1
        # else:
        #    sign_inv_vv = 1
        # sign_inv_vv = 1
        gradf_lat[:] = +gradf_lat / self.radius  # *sign_inv_vv
        gradf_lon[:] = +gradf_lon / self.radius
        # print('radius', self.radius)
        return gradf_lon, gradf_lat

    def spat3d_from_sh3d(self, f_lm_3d, f3d=None, dtype=np.float):
        nvert = f_lm_3d.shape[0]
        if f3d is None:
            f3d = np.empty([nvert, self.nlat, self.nlon], dtype)
        for iz in range(nvert):
            f2D = self.spat_from_sh(f_lm_3d[iz])
            f3d[iz] = f2D
        return f3d

    def sh3d_from_spat3d(self, f3d, f_lm3d=None, dtype=np.complex):
        nvert = f3d.shape[0]
        if f_lm3d is None:
            f_lm3d = np.empty([nvert, self.nlm], dtype)
        for iz in range(nvert):
            f_lm3d[iz] = self.sh_from_spat(f3d[iz])
        return f_lm3d

    # print functions...

    def print_info(self):
        print(self.produce_str_describing_oper())
        self.sh.print_info()
        print("")

    def print_array_sh(self, field_lm, name_field_lm, lmax_print=10):
        lmax_print = np.min([lmax_print, self.lmax])
        print(name_field_lm, "= ")
        for n in range(lmax_print + 1):
            print(f"n={n:2d} ", end="")
            for m in range(n + 1):
                temp_idx = m * self.lmax - (m + 1) * m / 2 + m + n
                sys.stdout.write("{:8.3g} ".format(np.abs(field_lm[temp_idx])))
            print("")

    # functions for the computation of spectra and co-spectra

    def sum_wavenumbers(self, field_lm):
        """Convenient function to look more like a pseudo-spectral Operators"""
        return field_lm.sum()

    def _spectrum_from_array_desh(self, array_desh):
        """Compute spectrum(l) from array_desh(ilm)"""
        spectrum = np.zeros(self.lmax + 1)
        for ilm in range(0, self.nlm):
            spectrum[self.l_idx[ilm]] += array_desh[ilm]
        return spectrum

    def _array_desh_from_sh(self, field_lm, key_field):
        """Compute the array_desh (density of energy) from an field_lm"""
        if key_field[:1] == "u" or key_field[:1] == "v":
            array_desh = abs(field_lm) ** 2 / 2.0
        elif (
            key_field[:1] == "T" or key_field[:2] == "ps" or key_field[:1] == "o"
        ):
            array_desh = abs(field_lm) ** 2 / 2.0
        elif key_field[:4] == "beta":
            array_desh = self.create_array_sh(0.0, float)
            array_desh = abs(field_lm) ** 2
        elif key_field[:2] == "uD" or key_field[:2] == "uR":
            array_desh = self.l2_idx * (abs(field_lm) ** 2) / 2
        elif key_field[:4] == "hdiv" or key_field[:4] == "hrot":
            array_desh = self.create_array_sh(0.0, float)
            COND = self.l2_idx > 0
            array_desh[COND] = (
                self.radius ** 2
                / self.l2_idx[COND]
                * abs(field_lm[COND]) ** 2
                / 2
            )
        else:
            raise ValueError("key_field is not correct")

        array_desh[self.m_idx > 0] = 2 * array_desh[self.m_idx > 0]
        return array_desh

    def spectrum_from_sh(self, field_lm, key_field):
        """ compute spectrum from field_lm"""
        array_desh = self._array_desh_from_sh(field_lm, key_field)
        spectrum = self._spectrum_from_array_desh(array_desh)
        return spectrum

    def cospectrum_from_2fieldssh(self, f_lm, g_lm):
        """ compute cospectrum(l) from f_lm(ilm) and g_lm(ilm)"""
        cospectrum = np.zeros(self.lmax + 1)

        array_desh = f_lm.conjugate() * g_lm + f_lm * g_lm.conjugate()
        array_desh = array_desh.real
        array_desh[self.m_idx == 0] = array_desh[self.m_idx == 0] / 2

        for ilm in range(self.nlm):
            cospectrum[self.l_idx[ilm]] += array_desh[ilm]
        return cospectrum

    def cospectrum_from_2vectorssh(self, f_lon_lm, f_lat_lm, g_lon_lm, g_lat_lm):
        """ compute cospectrum(l)..."""
        cospectrum = np.zeros(self.lmax + 1)

        array_desh = (
            f_lon_lm.conjugate() * g_lon_lm
            + f_lon_lm * g_lon_lm.conjugate()
            + f_lat_lm.conjugate() * g_lat_lm
            + f_lat_lm * g_lat_lm.conjugate()
        )
        array_desh = array_desh.real
        array_desh[self.m_idx == 0] = array_desh[self.m_idx == 0] / 2

        for ilm in range(self.nlm):
            cospectrum[self.l_idx[ilm]] += array_desh[ilm]
        return cospectrum

    def cospectrum_from_2fieldssh2(self, f_lm, g_lm):

        """ compute cospectrum(l)..."""
        cospectrum = np.zeros(self.lmax + 1)

        array_desh = f_lm.conjugate() * g_lm + f_lm * g_lm.conjugate()
        array_desh = array_desh.real
        array_desh[self.m_idx == 0] = array_desh[self.m_idx == 0] / 2

        array_desh2 = self.create_array_sh(0.0, float)
        COND = self.l2_idx > 0
        array_desh2[COND] = (
            self.radius ** 2 / self.l2_idx[COND] * array_desh[COND]
        )

        for ilm in range(self.nlm):
            cospectrum[self.l_idx[ilm]] += array_desh2[ilm]
        return cospectrum

    def cospectrum_from_2divrotsh(self, hdiva_lm, hrota_lm, hdivb_lm, hrotb_lm):
        """ compute cospectrum(l)..."""
        cospectrum = np.zeros(self.lmax + 1)

        array_desh = (
            hrota_lm.conjugate() * hrotb_lm + hdiva_lm.conjugate() * hdivb_lm
        )
        array_desh = array_desh.real
        array_desh[self.m_idx == 0] = array_desh[self.m_idx == 0] / 2

        array_desh2 = self.create_array_sh(0.0, float)
        COND = self.l2_idx > 0
        array_desh2[COND] = (
            self.radius ** 2 / self.l2_idx[COND] * array_desh[COND]
        )

        for ilm in range(self.nlm):
            cospectrum[self.l_idx[ilm]] += array_desh2[ilm]
        # print('cospe',cospectrum.shape)
        return cospectrum

    def dealiasing(self, field_lm):
        """Convenient function for fluidsim"""
        return field_lm


if __name__ == "__main__":
    obj = EasySHT()



"""Treat Matlab code to help the translation to Python code
===========================================================

This subpackage provides 2 modules to help a developer to translate a Matlab
code into a Python code. It is not a proper Matlab to Python translator. The
code produced is just more similar to Python than the original Matlab code.

.. autosummary::
   :toctree:

   cleanmat
   mat2wrongpy

"""

import os

import argparse


def get_index_closing_parenthesis(line, ind_start):
    if line[ind_start] != "(":
        raise ValueError("line[ind_start] != '('")

    in_parent = 1
    ind = ind_start
    while in_parent > 0:
        ind += 1
        if line[ind] == ")":
            in_parent -= 1
        elif line[ind] == "(":
            in_parent += 1

    return ind


# this can not be put at the top of the file and has to be put after the
# definition of get_index_closing_parenthesis
from . import cleanmat, mat2wrongpy


description = """

Utility to produce a strange code which is no longer Matlab and not yet Python.

"""


def main():

    parser = argparse.ArgumentParser(
        prog="fluidmat2py",
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "path",
        help="str indicating which file or directory has to be used.",
        type=str,
    )

    parser.add_argument(
        "-c", "--clean", help="Only clean the Matlab code.", action="store_true"
    )

    args = parser.parse_args()

    if not os.path.exists(args.path):
        raise ValueError("Path args.path does not exits.")

    if os.path.isdir(args.path):
        if args.clean:
            cleanmat.treat_matlab_directory(args.path)
        else:
            mat2wrongpy.treat_matlab_directory(args.path)
    else:
        if args.clean:
            code = cleanmat.modif_code(args.path)
        else:
            code = mat2wrongpy.create_py_code(args.path)

        print(code)
